from enum import Enum, auto

class TokenValues:
    cache = {}
    
    @staticmethod
    def get(token_class):
        assert token_class not in (TypedToken)
        if token_class not in TokenValues.cache:
            values = frozenset(token.value for token in token_class)
            TokenValues.cache[token_class] = values

        return TokenValues.cache[token_class]

class TokenClass:
    pass

class SimpleTokenClass(TokenClass):
    pass

class TypedToken(TokenClass, Enum):
    """
    Contains only a token type since the value is arbitrary
    """
    INT = auto()
    FLOAT = auto()
    STRING = auto()
    LITERAL_STRING = auto()
    # COORD = auto()
    SELECTOR_VARIABLE_SPECIFIER = auto()

class DelimiterToken(SimpleTokenClass, Enum):
    """
    Contains the type and value
    """
    QUOTE = '"'
    # ESCAPED_QUOTE = r'\"'
    COLON = ":"
    COMMA = ","
    AT = "@"
    EQUALS = "="
    EXCLAMATION_MARK = "!"
    RANGE = ".."
    SEMICOLON = ";"
    NUMBER_SIGN = "#"

    OPEN_PARENTHESES = "("
    CLOSE_PARENTHESES = ")"
    OPEN_SQUARE_BRACKET = "["
    CLOSE_SQUARE_BRACKET = "]"
    OPEN_CURLY_BRACKET = "{"
    CLOSE_CURLY_BRACKET = "}"

class WhitespaceToken(SimpleTokenClass, Enum):
    COMMENT = "#"
    INDENT = "    "
    DEDENT = "dedent"
    NEWLINE = "\n"
    EOF = "EOF (end of file)"

def test():
    print(DelimiterToken.COLON)
    print(repr(DelimiterToken.COLON))
    print(DelimiterToken.COLON in DelimiterToken)
    print(DelimiterToken.COLON in TypedToken)
    print(isinstance(DelimiterToken.COLON, TypedToken))
    print(isinstance(DelimiterToken, TypedToken))
    print(isinstance(DelimiterToken.COLON, SimpleTokenClass))
    print(isinstance(DelimiterToken, SimpleTokenClass))
    print(isinstance(DelimiterToken.COLON, TokenClass))
    print(isinstance(DelimiterToken, TokenClass))
    print(isinstance(TypedToken.INT, TypedToken))

    print(TokenValues.get(DelimiterToken))
    print(TokenValues.cache)
    print(TokenValues.get(DelimiterToken))

    print(TokenValues.get(DelimiterToken))

if __name__ == "__main__":
    test()



if __name__ == "__main__":
    import sys
    sys.path.append("..")
    del sys

from fenalib.assert_utils import assert_type
from fenalib.lexical_token import Token
from fenalib.nodes import CmdNode, StmtNode, ProgramNode

class TreePostfixTraversal:
    """
    Visitor from https://ruslanspivak.com/
    """


    def _visit(self, method_start, node, **kwargs):
        """
        Visits the specified node with the method starter

        Args:
            method_start (str): The starting part of the method name
                eg. method_start="visit", method_name=visit_<node>
            node (class type that inherits from Node)

        Returns:
            Whatever is gotten with the visitor method
        """
        assert_type(method_start, str)
        class_name = type(node).__name__
        method_name = f"{method_start}_{class_name}"
        visitor_method = getattr(self, method_name, "invalid")

        if visitor_method == "invalid":
            raise NotImplementedError(f"Invalid method: {method_name}")

        return visitor_method(node, **kwargs)

class NodeVisitor(TreePostfixTraversal):
    """
    Uses the visit method to visit any statment node
    """

    def visit(self, node, **kwargs):
        """
        Visits the specified node

        Returns:
            Whatever is gotten with the visitor method
        """
        # assert_type(node, StmtNode, Token, ProgramNode)
        return self._visit("visit", node, **kwargs)

class NodeBuilder(TreePostfixTraversal):
    """
    Uses the build method to traverse the tree to build fena commands
    """

    def build(self, node, **kwargs):
        """
        Builds the specified node

        Args:
            node (class type that inherits from Node)

        Returns:
            str: Whatever is gotten with the build method
        """
        # assert_type(node, CmdNode, Token, str)
        return self._visit("build", node, **kwargs)

    def iter_build(self, nodes, join_value=None):
        """
        Args:
            nodes (iterable object)
            join_value (str or None)

        Returns:
            generator (if join_value is None): generator to map all nodes to the build method
            str (if join_value is str): Full string of built objects from its nodes joined by the join_value
        """
        assert_type(join_value, str, optional=True)

        build_generator = map(self.build, nodes)
        if join_value is None:
            return build_generator

        return join_value.join(build_generator)




#!/usr/bin/env python3

"""
A butt load of code shamelessly taken from Ruslan Spivak's tutorial for making an interpreter:
    https://ruslanspivak.com/lsbasi-part1/

Pyexpander code shamelessly taken from:
    http://pyexpander.sourceforge.net/

Lots of config files shamelessly gotten from:
    https://github.com/PepijnMC/Minecraft
    https://github.com/Arcensoth/mcdata
"""

"""
Random stuff only for me honestly (vim session tab numbers):
    1: Python3/fena
        fena.py
    2: Fena/lib
        floo_event.py
        location.py
    3: floo_network/src
        main.py
        main.fena
    4: Python3/fena/fenalib/log
        after_pre_pyexpander.txt
        after_pyexpander.txt
        parsed_cmds.txt
    5: Fena.wiki
        Code-Style.md
        Code-Documentation.md
"""

import logging
import sys

import fenalib.logging_setup as logging_setup
logging_setup.setup_logging()

from fenalib.general import parse_text_from_args, get_args, semantic_version, public_version
# from fenalib.config_data import get_all_data


def main():
    # checks for python version (must be greater than or equal to 3.6)
    if not (sys.version_info.major >= 3 and sys.version_info.minor >= 6):
        raise RuntimeError("Python version must be 3.6 or greater")

    print(f"Fena: semantic_version={semantic_version}, public_version={public_version}")

    # gets all command line arguments
    args = get_args()

    #  overrides the version if necessary
    # if args.version is not None:
    #     get_all_data(args.version)

    with open(args.file_name) as file:
        text = file.read()

    # sets the file name for logging
    logging_setup.format_file_name(args.file_name)
    # parse_text(text, file_name, output_path, clean=args.clean, debug=args.debug)
    parse_text_from_args(text, args)


if __name__ == '__main__':
    # print("test")

    try:
        main()
    # pylint: disable=broad-except
    except Exception as e:
        logging.exception(e) # type: ignore




#!c:\python36\python.exe
# -*- coding: UTF-8 -*-
"""
expander.py: this is the pyexpander application.

NOTE: THIS IS A MODIFIED VERSION OF PYEXPANDER as of July 2018
License: https://bitbucket.org/goetzpf/pyexpander/src/b466de6fd801545650edfa790a18f022dc7e151a/LICENSE?at=default&fileviewer=file-view-default
Original: http://pyexpander.sourceforge.net/
"""

# pylint: disable=C0322,C0103,R0903

from optparse import OptionParser
#import string
import os.path
import sys

if __name__ == "__main__":
    sys.path.append("..")

import fena_pyexpander.lib as pyexpander

# version of the program:
__version__= "1.8.3" #VERSION#

assert __version__==pyexpander.__version__

def process_files(options,args):
    """process all the command line options."""
    my_globals={}
    if options.eval is not None:
        for expr in options.eval:
            # pylint: disable=W0122
            exec(expr, my_globals)
            # pylint: enable=W0122
    filelist= []
    if options.file is not None:
        filelist= options.file
    if len(args)>0: # extra arguments
        filelist.extend(args)
    if len(filelist)<=0:
        pyexpander.expandFile(filename=None,
                              external_definitions=my_globals,
                              allow_nobracket_vars=options.simple_vars,
                              auto_continuation=options.auto_continuation,
                              auto_indent=options.auto_indent,
                              include_paths=options.include,
                              auto_indent_python=options.auto_indent_python,
                              no_stdin_warning=options.no_stdin_msg)
    else:
        for f in filelist:
            # all files are expanded in a single scope:
            my_globals= \
                pyexpander.expandFile(filename=f,
                                      external_definitions=my_globals,
                                      allow_nobracket_vars=options.simple_vars,
                                      auto_continuation=options.auto_continuation,
                                      auto_indent=options.auto_indent,
                                      auto_indent_python=options.auto_indent_python,
                                      include_paths=options.include)

def script_shortname():
    """return the name of this script without a path component."""
    return os.path.basename(sys.argv[0])

def print_summary():
    """print a short summary of the scripts function."""
    print(("%-20s: a powerful macro expension language "+\
           "based on python ...\n") % script_shortname())

def main():
    """The main function.

    parse the command-line options and perform the command
    """
    # command-line options and command-line help:
    usage = "usage: %prog [options] {files}"

    parser = OptionParser(usage=usage,
                          version="%%prog %s" % __version__,
                          description="expands macros in a file "+\
                                      "with pyexpander.")

    parser.add_option("--summary",
                      action="store_true",
                      help="Print a summary of the function of the program.",
                     )
    parser.add_option("-f", "--file",
                      action="append",
                      type="string",
                      help="Specify a FILE to process. This "
                           "option may be used more than once "
                           "to process more than one file but note "
                           "than this option is not really needed. "
                           "Files can also be specified directly after "
                           "the other command line options. If not given, "
                           "the program gets it's input from stdin.",
                      metavar="FILE"
                     )
    parser.add_option("--eval",
                      action="append",
                      type="string",
                      help="Evaluate PYTHONEXPRESSION in global context.",
                      metavar="PYTHONEXPRESSION"
                     )
    parser.add_option("-I", "--include",
                      action="append",
                      type="string",
                      help="Add PATH to the list of include paths.",
                      metavar="PATH"
                     )
    parser.add_option("-s", "--simple-vars",
                      action="store_true",
                      help="Allow variables without brackets.",
                     )
    parser.add_option("-a", "--auto-continuation",
                      action="store_true",
                      help="Assume '\' at the end of lines with commands",
                     )
    parser.add_option("-i", "--auto-indent",
                      action="store_true",
                      help="Automatically indent macros.",
                     )
    parser.add_option("-p", "--auto-indent-python",
                      action="store_true",
                      help="Automatically indent python output.",
                     )
    parser.add_option("--no-stdin-msg",
                      action="store_true",
                      help= "Do not print a message on stderr when the "
                            "program is reading it's input from stdin."
                     )

    # x= sys.argv
    (options, args) = parser.parse_args()
    # options: the options-object
    # args: list of left-over args

    if options.summary:
        print_summary()
        sys.exit(0)

    process_files(options,args)
    sys.exit(0)

if __name__ == "__main__":
    main()




from fenalib.assert_utils import assert_type, assert_list_types
from fenalib.repr_utils import addrepr

@addrepr
class McFunction:
    """
    Args:
        full_path (str)
        mfunc_name (str)
        debug (bool)

    Attributes:
        full_path (str): The full path to the mcfunction file
        mfunc_name (str)
        debug (bool): Whether to debug or not with the debug command line option
        commands (list or tuple): The full sequence of strings in an mcfunction
    """
    def __init__(self, mfunc_name, full_path, debug):
        assert_type(mfunc_name, str)
        assert_type(full_path, str)
        assert_type(debug, bool)

        self.mfunc_name = mfunc_name
        self.full_path = full_path
        self.debug = debug
        self.commands = []
        self._finalized = False

    @property
    def finalized(self):
        return self._finalized

    def add_command(self, command):
        """
        Adds the given command to the commands list

        Args:
            command (str)
        """
        assert_type(command, str)
        assert not self._finalized, "Cannot add a command if finalized"
        self.commands.append(command)

    def finalize(self):
        """
        Converts all mutable attributes into immutable attributes
        """
        assert_type(self.commands, list)
        assert not self._finalized
        self.commands = tuple(self.commands)
        self._finalized = True

    def __str__(self):
        return f"McFunction[{self.mfunc_name}]"



import logging

if __name__ == "__main__":
    import sys
    sys.path.append("..")
    del sys

from fenalib.assert_utils import assert_type
from fenalib.token_position import TokenPosition
from fenalib.token_classes import TypedToken, DelimiterToken, WhitespaceToken
from fenalib.token_classes import SimpleTokenClass, TokenClass
from fenalib.coord_utils import is_coord
from fenalib.config_data import ConfigData

"""
Note that this file is named "lexical_token" because if it was named "token",
it would interfere with the token used from the default logging module
"""

class Token:
    # typed_tokens = frozenset(TypedToken)
    # simple_token_types = TokenTypes.get(DelimiterToken, WhitespaceToken)
    # all_token_types = typed_tokens | simple_token_types

    config_data = ConfigData()

    def __init__(self, position, token_type, value=None):
        """
        Args:
            position (TokenPosition): position inside the file formatted as (row, column)
            type (TypedToken, SimpleToken, WhitespaceToken, StatementToken) type of the token
            value (any): an optional custom value
            replacement (any): an optional custom value to be set as a replacement of the token type
                - this will be used only when building instead of returning the value
        """
        assert_type(position, TokenPosition)
        assert_type(token_type, TokenClass)
        assert_type(value, str, optional=True)

        self.position = position
        self.token_type = token_type
        self.value = value
        self.replacement = None

        if self.value is None:
            assert_type(token_type, SimpleTokenClass)
            self.value = self.token_type.value

    def matches(self, token_type, value=None, values=None):
        """
        returns whether the token matches the given type and/or value

        Args:
            type (any token type)
            value (optional, any type)

        Returns:
            bool: Whether the type matches the given type
        """
        assert_type(token_type, TokenClass)
        assert not (value is not None and values is not None)
        if value is not None:
            return (self.token_type == token_type) and self.value == value

        if values is not None:
            return (self.token_type == token_type) and self.value in values

        return self.token_type == token_type

    def matches_any_of(self, *types):
        """
        returns whether the token matches any one of the types

        Args:
            types (any token type): any number of types to compare the token with

        Returns:
            bool: Whether the token matches any of the provided types
        """
        for token_type in types:
            if self.matches(token_type):
                return True
        return False

    def __str__(self):
        return f"Token[{self.value!r} ({str(self.token_type)}) at {self.position}]"

    def __repr__(self):
        if self.replacement is None:
            return f'Token[{self.position!r}: type={self.token_type!r}, value={self.value!r}]'
        return f'Token[{self.position!r}: type={self.token_type!r}, value={self.value!r}, replacement={self.replacement!r}]'


def test():
    from token_position import TokenPosition
    # from token_classes import SimpleToken

    token_pos = TokenPosition(row=5, column=2, char_pos=167)

    # token = Token(token_pos, SimpleToken.PLUS)
    token = Token(token_pos, DelimiterToken.EQUALS)
    print(token)
    print(repr(token))

    # print(token.type in ALL_TOKEN_TYPES)
    # print(token.type in SimpleToken)

    # error since it requires a value
    # integer = Token(token_pos, TypedToken.INT)
    integer = Token(token_pos, TypedToken.INT, 26)
    print(repr(integer))

def test_docs():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    test_docs()
    test()



from abc import ABC

if __name__ == "__main__":
    import sys
    sys.path.append("..")
    del sys

from fenalib.config_data import ConfigData
from fenalib.repr_utils import addrepr
from fenalib.assert_utils import assert_type, assert_list_types, assert_tuple_types
from fenalib.lexical_token import Token

config_data = ConfigData()

"""
Backup module when changing to using NamedTuple
"""

@addrepr
class Node(ABC):
    """
    All nodes specified below will inherit from this parent class

    Adds a simple __repr__ method to all of its classes to display its own variables
    """
    pass

class ProgramNode(Node):
    """
    Holds all statements found inside the global scope of the program

    Attributes:
        statement_nodes (list of nodes inherited from StmtNode)
    """
    def __init__(self, statement_nodes):
        assert_type(statement_nodes, list)
        assert_list_types(statement_nodes, StmtNode)
        self.statement_nodes = statement_nodes


class StmtNode(Node, ABC):
    """
    General node to be inherited from all specialized statement nodes
    """
    pass

class McFunctionNode(StmtNode):
    """
    Holds a single mcfunction with its full path and all fena command nodes

    Attributes:
        name (Token): The mcfunction name
        command_nodes (list of FenaCmdNode objects)
    """
    def __init__(self, name, command_nodes):
        assert_type(name, Token)
        assert_list_types(command_nodes, CmdNode)
        self.name = name
        self.command_nodes = command_nodes

class FolderNode(StmtNode):
    """
    Holds all statements found inside a folder node

    Attributes:
        folder (Token)
        statement_nodes (list of nodes inherited from StmtNode)
    """
    def __init__(self, folder, statement_nodes):
        assert_type(folder, Token)
        assert_list_types(statement_nodes, StmtNode)
        self.folder = folder
        self.statement_nodes = statement_nodes

class VarSetNode(StmtNode):
    """
    Holds all variables defined for the file

    Attributes:
        variable (Token)
        value (Token)
    """
    def __init__(self, variable, value):
        assert_type(variable, Token)
        assert_type(value, Token)
        self.variable = variable
        self.value = value


class CmdNode(Node, ABC):
    """
    General node to be inherited from all nodes relating to commands
    """
    pass

class FenaCmdNode(CmdNode):
    """
    Holds all command segments for a command node

    Args:
        cmd_segment_nodes (list of MainCmdNode objects)
            - should contain one of {ExecuteCmdNode, ScoreboardCmdNode, FunctionCmdNode, SimpleCmdNode}
    """
    def __init__(self, cmd_segment_nodes):
        assert_list_types(cmd_segment_nodes, MainCmdNode)
        self.cmd_segment_nodes = cmd_segment_nodes

class MainCmdNode(CmdNode, ABC):
    """
    General node to be inherited from all children nodes to the FenaCmdNode
        - Nodes that a command is made up of in the most general context
        - eg. ExecuteCmdNode, SimpleCmdNode
    """
    pass

class ExecuteCmdNode(MainCmdNode):
    """
    Attributes:
        sub_cmd_nodes (list of ExecuteSubArgNode, SelectorNode and Vec3Node objects or list of ExecuteSubLegacyArg in 1.12)
    """
    def __init__(self, sub_cmd_nodes):
        if config_data.version == "1.12":
            assert_list_types(sub_cmd_nodes, ExecuteSubLegacyArg)
        else:
            assert_list_types(sub_cmd_nodes, ExecuteSubArgNode, SelectorNode, Vec3Node)
        self.sub_cmd_nodes = sub_cmd_nodes

class ExecuteSubLegacyArg(MainCmdNode):
    """
    Execute node for all versions under 1.12 inclusive

    Attributes:
        selector (SelectorNode)
        coords (Vec3Node or None)
        sub_if (list of ExecuteSubIfBlockArg objects)
    """
    def __init__(self, selector, coords=None, sub_if=None):
        assert_type(selector, SelectorNode)
        assert_type(coords, Vec3Node, optional=True)
        assert_list_types(sub_if, ExecuteSubIfBlockArg)
        self.selector = selector
        self.coords = coords
        self.sub_if = sub_if

class ExecuteSubArgNode(CmdNode, ABC):
    pass

class ExecuteSubArgSelectorNode(ExecuteSubArgNode, ABC):
    """
    Attributes:
        selector (SelectorNode)
    """
    def __init__(self, selector):
        assert_type(selector, SelectorNode)
        self.selector = selector

class ExecuteSubArgVec2Node(ExecuteSubArgNode, ABC):
    """
    Attributes:
        vec2 (Vec2Node)
    """
    def __init__(self, vec2):
        assert_type(vec2, Vec2Node)
        self.vec2 = vec2

class ExecuteSubArgVec3Node(ExecuteSubArgNode, ABC):
    """
    Attributes:
        vec3 (Vec3Node)
    """
    def __init__(self, vec3):
        assert_type(vec3, Vec3Node)
        self.vec3 = vec3

class ExecuteSubAsArg(ExecuteSubArgSelectorNode):
    pass

class ExecuteSubPosVec3Arg(ExecuteSubArgVec3Node):
    pass

class ExecuteSubPosSelectorArg(ExecuteSubArgSelectorNode):
    pass

class ExecuteSubAtAnchorArg(ExecuteSubArgNode):
    """
    Attributes:
        anchor (Token)
    """
    def __init__(self, anchor):
        assert_type(anchor, Token)
        self.anchor = anchor

class ExecuteSubAtSelectorArg(ExecuteSubArgSelectorNode):
    pass

class ExecuteSubAtCoordsArg(ExecuteSubArgNode):
    """
    Attributes:
        vec3 (Vec3Node)
        vec2 (Vec2Node)
    """
    def __init__(self, vec3, vec2):
        assert_type(vec3, Vec3Node)
        assert_type(vec2, Vec2Node)
        self.vec3 = vec3
        self.vec2 = vec2

class ExecuteSubAtAxesArg(ExecuteSubArgNode):
    """
    Attributes:
        axes (Token)
    """
    def __init__(self, axes):
        assert_type(axes, Token)
        self.axes = axes

class ExecuteSubFacingVec3Arg(ExecuteSubArgVec3Node):
    """
    Attributes:
        vec3 (Vec3Node)
    """
    pass

class ExecuteSubFacingSelectorArg(ExecuteSubArgNode):
    """
    Attributes:
        selector (SelectorNode)
        anchor (Token or None)
    """
    def __init__(self, selector, anchor=None):
        assert_type(selector, SelectorNode)
        assert_type(anchor, Token, optional=True)
        self.selector = selector
        self.anchor = anchor

class ExecuteSubRotSelectorArg(ExecuteSubArgSelectorNode):
    pass

class ExecuteSubRotVec2Arg(ExecuteSubArgVec2Node):
    """
    Attributes:
        vec2 (Vec2Node)
    """
    pass

class ExecuteSubAnchorArg(ExecuteSubArgNode):
    """
    Attributes:
        axes (Token)
    """
    def __init__(self, axes):
        assert_type(axes, Token)
        self.axes = axes

class ExecuteSubInArg(ExecuteSubArgNode):
    """
    Attributes:
        dimension (Token)
    """
    def __init__(self, dimension):
        assert_type(dimension, Token)
        self.dimension = dimension

class ExecuteSubAstArg(ExecuteSubArgSelectorNode):
    pass

class ExecuteSubIfArg(ExecuteSubArgNode):
    """
    Args:
        negated (bool): Whether the statement argument should be negated or not when built

    Attributes:
        sub_cmd (str): Either "if" if negated is false, or "unless" if negated is true
    """
    def __init__(self, negated):
        assert_type(negated, bool)
        self.sub_cmd = ("unless" if negated else "if")

class ExecuteSubIfSelectorArg(ExecuteSubIfArg):
    """
    Attributes:
        selector (SelectorNode)
    """
    def __init__(self, selector, negated=False):
        assert_type(selector, SelectorNode)
        super().__init__(negated)
        self.selector = selector

class ExecuteSubIfBlockArg(ExecuteSubIfArg):
    """
    Attributes:
        block (BlockNode)
        coords (Vec3Node or None)
    """
    def __init__(self, block, coords=None, negated=False):
        assert_type(block, BlockNode)
        assert_type(coords, Vec3Node, optional=True)

        super().__init__(negated)
        self.block = block
        self.coords = coords

class ExecuteSubIfBlocksArg(ExecuteSubIfArg):
    """
    Attributes:
        coords1 (Vec3Node)
        coords2 (Vec3Node)
        coords3 (Vec3Node)
        option (Token or None)
    """
    def __init__(self, coords1, coords2, coords3, option=None, negated=False):
        assert_type(coords1, Vec3Node)
        assert_type(coords2, Vec3Node)
        assert_type(coords3, Vec3Node)
        assert_type(option, Token, optional=True)

        super().__init__(negated)
        self.coords1 = coords1
        self.coords2 = coords2
        self.coords3 = coords3
        self.option = option

class ExecuteSubIfCompareEntityArg(ExecuteSubIfArg):
    """
    Attributes:
        target (SelectorNode or Token)
        objective (Token)
        operator (Token)
        target_get (SelectorNode)
        objective_get (Token)
    """
    def __init__(self, target, objective, operator, target_get, objective_get, negated=False):
        assert_type(target, SelectorNode, Token)
        assert_type(objective, Token)
        assert_type(operator, Token)
        assert_type(target_get, SelectorNode)
        assert_type(objective_get, Token)

        super().__init__(negated)
        self.target = target
        self.objective = objective
        self.operator = operator
        self.target_get = target_get
        self.objective_get = objective_get

class ExecuteSubIfCompareIntArg(ExecuteSubIfArg):
    """
    Attributes:
        target (SelectorNode or Token)
        objective (Token)
        operator (Token)
        value (Token)

    Usage:
        if(target objective == int) -> if score target objective matches int
        if(target objective < int) -> if score target objective matches ..(int-1)
        if(target objective <= int) -> if score target objective matches ..(int)
        if(target objective > int) -> if score target objective matches (int+1)..
        if(target objective >= int) -> if score target objective matches (int)..
    """
    def __init__(self, target, objective, operator, value, negated=False):
        assert_type(target, SelectorNode, Token)
        assert_type(objective, Token)
        assert_type(operator, Token)
        assert_type(value, Token)

        super().__init__(negated)
        self.target = target
        self.objective = objective
        self.operator = operator
        self.value = value

class ExecuteSubIfRangeArg(ExecuteSubIfArg):
    """
    Attributes:
        target (SelectorNode or Token)
        objective (Token)
        int_range (Token)

    Usage:
        if(target objective in int..int) -> if score target objective matches int..int
    """
    def __init__(self, target, objective, int_range, negated=False):
        assert_type(target, SelectorNode, Token)
        assert_type(objective, Token)
        assert_type(int_range, Token)

        super().__init__(negated)
        self.target = target
        self.objective = objective
        self.int_range = int_range

class ExecuteSubStoreArg(ExecuteSubArgNode):
    def __init__(self, store_type):
        assert_type(store_type, str)
        self.store_type = store_type

class ExecuteSubStoreSelectorDataArg(ExecuteSubStoreArg):
    """
    Attributes:
        selector (SelectorNode)
        data_path (DataPathNode)
        scale (Token)
        data_type (Token)
    """
    def __init__(self, store_type, selector, data_path, scale, data_type=None):
        super().__init__(store_type)
        assert_type(selector, SelectorNode)
        assert_type(data_path, DataPathNode)
        assert_type(scale, Token)
        assert_type(data_type, Token, optional=True)

        self.selector = selector
        self.data_path = data_path
        self.scale = scale
        self.data_type = data_type

class ExecuteSubStoreVec3DataArg(ExecuteSubStoreArg):
    """
    Attributes:
        vec3 (Vec3Node)
        data_path (DataPathNode)
        scale (Token)
        data_type (Token)
    """
    def __init__(self, store_type, vec3, data_path, scale, data_type=None):
        super().__init__(store_type)
        assert_type(vec3, SelectorNode, Vec3Node)
        assert_type(data_path, DataPathNode)
        assert_type(scale, Token)
        assert_type(data_type, Token, optional=True)

        self.vec3 = vec3
        self.data_path = data_path
        self.scale = scale
        self.data_type = data_type

class ExecuteSubStoreScoreArg(ExecuteSubStoreArg):
    """
    Attributes:
        target (SelectorNode or Token)
        objective (Token)
    """
    def __init__(self, store_type, target, objective):
        super().__init__(store_type)
        assert_type(target, SelectorNode, Token)
        assert_type(objective, Token)
        self.target = target
        self.objective = objective

class ExecuteSubStoreBossbarArg(ExecuteSubStoreArg):
    """
    Attributes:
        bossbar_id (NamespaceIdNode)
        sub_cmd (Token)
    """
    def __init__(self, store_type, bossbar_id, sub_cmd):
        super().__init__(store_type)
        assert_type(bossbar_id, NamespaceIdNode)
        assert_type(sub_cmd, Token)
        self.bossbar_id = bossbar_id
        self.sub_cmd = sub_cmd

class ExecuteSubDataArg(ExecuteSubStoreArg):
    pass

class ExecuteSubScoreArg(ExecuteSubStoreArg):
    pass

class ExecuteSubBossbarArg(ExecuteSubStoreArg):
    pass


class ScoreboardCmdMathNode(MainCmdNode):
    """
    Args:
        target (SelectorNode or Token)
        objective (Token)
        operator (Token)
        target_get (SelectorNode or Token)
        objective_get (Token or None)
    """

    def __init__(self, target, objective, operator, target_get, objective_get=None):
        assert_type(target, SelectorNode, Token)
        assert_type(objective, Token)
        assert_type(operator, Token)
        assert_type(target_get, SelectorNode, Token)
        assert_type(objective_get, Token, optional=True)

        self.target = target
        self.objective = objective
        self.operator = operator
        self.target_get = target_get
        self.objective_get = objective_get

class ScoreboardCmdMathValueNode(MainCmdNode):
    """
    Args:
        target (SelectorNode or Token)
        objective (Token)
        operator (Token)
        value (Token)
        nbt (NbtObjectNode or None)
    """
    def __init__(self, target, objective, operator, value, nbt=None):
        assert_type(target, SelectorNode, Token)
        assert_type(objective, Token)
        assert_type(operator, Token)
        assert_type(value, Token)
        assert_type(nbt, NbtObjectNode, optional=True)

        self.target = target
        self.objective = objective
        self.operator = operator
        self.value = value
        self.nbt = nbt

class ScoreboardCmdSpecialNode(MainCmdNode):
    """
    Args:
        target (SelectorNode or Token)
        sub_cmd (Token)
        objective (Token)
    """
    def __init__(self, target, sub_cmd, objective):
        assert_type(target, SelectorNode, Token)
        assert_type(sub_cmd, Token)
        assert_type(objective, Token)

        self.target = target
        self.sub_cmd = sub_cmd
        self.objective = objective

class SimpleCmdNode(MainCmdNode):
    """
    Attributes:
        nodes (list of Token, SelectorNode, JsonObjectNode and NbtObjectNode objects)
    """
    def __init__(self, tokens):
        assert_type(tokens, list)
        assert_list_types(tokens, Token, SelectorNode, JsonObjectNode, NbtObjectNode, NamespaceIdNode)
        self.tokens = tokens

class BossbarCmdNode(MainCmdNode, ABC):
    pass

class BossbarAddNode(BossbarCmdNode):
    """
    Attributes:
        bossbar_id (NamespaceIdNode)
        json (JsonObjectNode)
    """
    def __init__(self, bossbar_id, json):
        assert_type(bossbar_id, NamespaceIdNode)
        assert_type(json, JsonObjectNode)
        self.bossbar_id = bossbar_id
        self.json = json

class BossbarRemoveNode(BossbarCmdNode):
    """
    Attributes:
        bossbar_id (NamespaceIdNode)
    """
    def __init__(self, bossbar_id):
        assert_type(bossbar_id, NamespaceIdNode)
        self.bossbar_id = bossbar_id

class BossbarGetNode(BossbarCmdNode):
    """
    Attributes:
        bossbar_id (NamespaceIdNode)
        sub_cmd (Token)
    """
    def __init__(self, bossbar_id, sub_cmd):
        assert_type(bossbar_id, NamespaceIdNode)
        assert_type(sub_cmd, Token)
        self.bossbar_id = bossbar_id
        self.sub_cmd = sub_cmd

class BossbarSetNode(BossbarCmdNode):
    """
    Attributes:
        bossbar_id (NamespaceIdNode)
        arg (Token)
        arg_value (Token, JsonObjectNode, SelectorNode)
    """
    def __init__(self, bossbar_id, arg, arg_value):
        assert_type(bossbar_id, NamespaceIdNode)
        assert_type(arg, Token)
        assert_type(arg_value, Token, JsonObjectNode, SelectorNode)
        self.bossbar_id = bossbar_id
        self.arg = arg
        self.arg_value = arg_value

class DataCmdNode(MainCmdNode):
    pass

class DataGetNode(DataCmdNode):
    """
    Attributes:
        entity_vec3 (SelectorNode, Vec3Node)
        data_path (DataPathNode or None)
        scale (Token or None)
    """
    def __init__(self, entity_vec3, data_path=None, scale=None):
        assert_type(entity_vec3, SelectorNode, Vec3Node)
        assert_type(data_path, DataPathNode, optional=True)
        assert_type(scale, Token, optional=True)
        self.entity_vec3 = entity_vec3
        self.data_path = data_path
        self.scale = scale

class DataMergeNode(DataCmdNode):
    """
    Attributes:
        entity_vec3 (SelectorNode, Vec3Node)
        nbt (NbtObjectNode)
    """
    def __init__(self, entity_vec3, nbt):
        assert_type(entity_vec3, SelectorNode, Vec3Node)
        assert_type(nbt, NbtObjectNode)
        self.entity_vec3 = entity_vec3
        self.nbt = nbt

class DataRemoveNode(DataCmdNode):
    """
    Attributes:
        entity_vec3 (SelectorNode, Vec3Node)
        data_path (DataPathNode)
    """
    def __init__(self, entity_vec3, data_path):
        assert_type(entity_vec3, SelectorNode, Vec3Node)
        assert_type(data_path, DataPathNode)
        self.entity_vec3 = entity_vec3
        self.data_path = data_path


class EffectCmdNode(MainCmdNode):
    pass

class EffectClearNode(EffectCmdNode):
    """
    Attributes:
        selector (SelectorNode)
        effect_id (Token or None)
    """
    def __init__(self, selector, effect_id=None):
        assert_type(selector, SelectorNode)
        assert_type(effect_id, Token, optional=True)
        self.selector = selector
        self.effect_id = effect_id

class EffectGiveNode(EffectCmdNode):
    """
    Attributes:
        selector (SelectorNode)
        effect_id (Token)
        duration (Token or None)
        level (Token or None)
        hide_particles (bool)
    """
    def __init__(self, selector, effect_id, duration=None, level=None, hide_particles=True):
        assert_type(selector, SelectorNode)
        assert_type(effect_id, Token)
        assert_type(duration, Token, optional=True)
        assert_type(level, Token, optional=True)
        assert_type(hide_particles, bool)
        self.selector = selector
        self.effect_id = effect_id
        self.duration = duration
        self.level = level
        self.hide_particles = hide_particles


class FunctionCmdNode(MainCmdNode):
    """
    Attributes:
        function_id (NamespaceIdNode)
        sub_arg (Token or None)
        selector (SelectorNode or None)
    """
    def __init__(self, function_id, sub_arg=None, selector=None):
        assert_type(function_id, NamespaceIdNode)
        assert_type(sub_arg, Token, optional=True)
        assert_type(selector, SelectorNode, optional=True)
        self.function_id = function_id
        self.sub_arg = sub_arg
        self.selector = selector

class ItemCmdNode(MainCmdNode):
    pass

class ItemGiveNode(ItemCmdNode):
    """
    Attributes:
        selector (SelectorNode)
        item (ItemNode)
        count (Token or None)
    """
    def __init__(self, selector, item, count=None):
        assert_type(selector, SelectorNode)
        assert_type(item, ItemNode)
        assert_type(count, Token, optional=True)
        self.selector = selector
        self.item = item
        self.count = count

class ItemClearNode(ItemCmdNode):
    """
    Attributes:
        selector (SelectorNode)
        item (ItemNode or Token)
        count (Token or None)
    """
    def __init__(self, selector, item, count=None):
        assert_type(selector, SelectorNode)
        assert_type(item, ItemNode, Token)
        assert_type(count, Token, optional=True)
        self.selector = selector
        self.item = item
        self.count = count

class ItemReplaceEntityNode(ItemCmdNode):
    """
    Attributes:
        selector (SelectorNode)
        slot (Token)
        item (ItemNode)
        count (Token or None)
    """
    def __init__(self, selector, slot, item, count=None):
        assert_type(selector, SelectorNode)
        assert_type(slot, Token)
        assert_type(item, ItemNode)
        assert_type(count, Token, optional=True)
        self.selector = selector
        self.slot = slot
        self.item = item
        self.count = count

class ItemReplaceBlockNode(ItemCmdNode):
    """
    Attributes:
        vec3 (Vec3Node)
        slot (Token)
        item (ItemNode)
        count (Token or None)
    """
    def __init__(self, vec3, slot, item, count=None):
        assert_type(vec3, Vec3Node)
        assert_type(slot, Token)
        assert_type(item, ItemNode)
        assert_type(count, Token, optional=True)
        self.vec3 = vec3
        self.slot = slot
        self.item = item
        self.count = count

class ItemNode(CmdNode):
    """
    Attributes:
        item_id (Token)
        damage (Token or None)
        nbt (NbtObjectNode or None)
    """
    def __init__(self, item_id, damage=None, nbt=None):
        assert_type(item_id, Token)
        assert_type(damage, Token, optional=True)
        assert_type(nbt, NbtObjectNode, optional=True)
        self.item_id = item_id
        self.damage = damage
        self.nbt = nbt


class ObjectiveCmdNode(MainCmdNode):
    pass

class ObjectiveAddNode(ObjectiveCmdNode):
    """
    Attributes:
        objective (Token)
        criteria (Token)
        display_name (List[Token])
    """
    def __init__(self, objective, criteria, display_name):
        assert_type(objective, Token)
        assert_type(criteria, Token)
        assert_list_types(display_name, Token)
        self.objective = objective
        self.criteria = criteria
        self.display_name = display_name

class ObjectiveRemoveNode(ObjectiveCmdNode):
    """
    Attributes:
        objective (Token)
    """
    def __init__(self, objective):
        assert_type(objective, Token)
        self.objective = objective

class ObjectiveSetdisplayNode(ObjectiveCmdNode):
    """
    Attributes:
        slot (Token)
        objective (Token or None)
    """
    def __init__(self, slot, objective=None):
        assert_type(slot, Token)
        assert_type(objective, Token, optional=True)
        self.slot = slot
        self.objective = objective


class TagCmdNode(MainCmdNode):
    pass

class TagAddNode(TagCmdNode):
    """
    Attributes:
        selector (SelectorNode)
        tag (Token)
        nbt (NbtObjectNode or None)
    """
    def __init__(self, selector, tag, nbt=None):
        assert_type(selector, SelectorNode)
        assert_type(tag, Token)
        assert_type(nbt, NbtObjectNode, optional=True)
        self.selector = selector
        self.tag = tag
        self.nbt = nbt

class TagRemoveNode(TagCmdNode):
    """
    Attributes:
        selector (SelectorNode)
        tag (Token)
        nbt (NbtObjectNode or None)
    """
    def __init__(self, selector, tag, nbt=None):
        assert_type(selector, SelectorNode)
        assert_type(tag, Token)
        assert_type(nbt, NbtObjectNode, optional=True)
        self.selector = selector
        self.tag = tag
        self.nbt = nbt


class TeamCmdNode(MainCmdNode, ABC):
    pass

class TeamAddNode(TeamCmdNode):
    """
    team_add ::= "add" && STR && (STR)*

    Attributes:
        team_name (Token)
        display_name (list of Token objects)
    """
    def __init__(self, team_name, display_name):
        assert_type(team_name, Token)
        assert_list_types(display_name, Token)
        self.team_name = team_name
        self.display_name = display_name

class TeamJoinNode(TeamCmdNode):
    """
    team_join ::= STR && "+=" && target

    Attributes:
        team_name (Token)
        target (SelectorNode or Token)
    """
    def __init__(self, team_name, target):
        assert_type(team_name, Token)
        assert_type(target, Token, SelectorNode)
        self.team_name = team_name
        self.target = target

class TeamLeaveNode(TeamCmdNode):
    """
    team_leave ::= "leave" && target

    Attributes:
        target (SelectorNode or Token)
    """
    def __init__(self, target):
        assert_type(target, SelectorNode, Token)
        self.target = target

class TeamEmptyNode(TeamCmdNode):
    """
    team_empty ::= "empty" && STR

    Attributes:
        team_name (Token)
    """
    def __init__(self, team_name):
        assert_type(team_name, Token)
        self.team_name = team_name

class TeamOptionNode(TeamCmdNode):
    """
    Attributes:
        team_name (Token)
        option (Token)
        value (Token, JsonObjectNode)
    """
    def __init__(self, team_name, option, value):
        assert_type(team_name, Token)
        assert_type(option, Token)
        assert_type(value, Token, JsonObjectNode)
        self.team_name = team_name
        self.option = option
        self.value = value

class TeamRemoveNode(TeamCmdNode):
    """
    Attributes:
        team_name (Token)
    """
    def __init__(self, team_name):
        assert_type(team_name, Token)
        self.team_name = team_name

class XpCmdNode(MainCmdNode, ABC):
    pass

class XpMathNode(XpCmdNode):
    """
    Attributes:
        selector (SelectorNode)
        operator (Token)
        value (Token)
        sub_cmd (Token or None)
    """
    def __init__(self, selector, operator, value, sub_cmd=None):
        assert_type(selector, SelectorNode)
        assert_type(operator, Token)
        assert_type(value, Token)
        assert_type(sub_cmd, Token, optional=True)
        self.selector = selector
        self.operator = operator
        self.value = value
        self.sub_cmd = sub_cmd

class XpGetNode(XpCmdNode):
    """
    Attributes:
        selector (SelectorNode)
        sub_cmd (Token)
    """
    def __init__(self, selector, sub_cmd):
        assert_type(selector, SelectorNode)
        assert_type(sub_cmd, Token)
        self.selector = selector
        self.sub_cmd = sub_cmd

class SelectorNode(CmdNode):
    """
    Attributes:
        selector_var (SelectorVarNode)
        selector_args (SelectorArgsNode or None)
    """
    def __init__(self, selector_var, selector_args=None):
        assert_type(selector_var, SelectorVarNode)
        assert_type(selector_args, SelectorArgsNode, optional=True)

        self.selector_var = selector_var
        self.selector_args = selector_args

class SelectorVarNode(CmdNode):
    """
    Attributes:
        selector_var_specifier (Token)
    """
    def __init__(self, selector_var_specifier):
        assert_type(selector_var_specifier, Token)
        self.selector_var_specifier = selector_var_specifier

class SelectorArgsNode(CmdNode):
    """
    Attributes:
        default_args (list of SelectorDefaultArgNode objects)
        score_args (SelectorScoreArgsNode)
        tag_args (SelectorTagArgsNode)
        nbt_args (SelectorNbtArgsNode)
        advancement_args (SelectorAdvancementGroupArgNode)
    """
    def __init__(self, default_args, score_args, tag_args, nbt_args, advancement_args):
        assert_list_types(default_args, SelectorDefaultArgNode, duplicate_key=lambda x: x.arg.value)
        assert_type(score_args, SelectorScoreArgsNode)
        assert_type(tag_args, SelectorTagArgsNode)
        assert_type(nbt_args, SelectorNbtArgsNode)
        assert_type(advancement_args, SelectorAdvancementGroupArgNode)

        self.default_args = default_args
        self.score_args = score_args
        self.tag_args = tag_args
        self.nbt_args = nbt_args
        self.advancement_args = advancement_args

class SelectorScoreArgsNode(CmdNode):
    """
    Attributes:
        score_args (list of SelectorScoreArgNode objects)
    """
    def __init__(self, score_args):
        assert_list_types(score_args, SelectorScoreArgNode, duplicate_key=lambda x: x.objective.value)
        self.score_args = score_args

class SelectorScoreArgNode(CmdNode):
    """
    Attributes:
        objective (Token)
        value (IntRangeNode, Token)
    """
    def __init__(self, objective, value):
        assert_type(objective, Token)
        assert_type(value, IntRangeNode, Token)
        self.objective = objective
        self.value = value

class SelectorDefaultArgNode(CmdNode):
    """
    Attributes:
        arg (Token)
        arg_value (SelectorDefaultArgValueNode, SelectorDefaultGroupArgValueNode)
    """
    def __init__(self, arg, arg_value):
        assert_type(arg, Token)
        assert_type(arg_value, (SelectorDefaultArgValueNode, SelectorDefaultGroupArgValueNode))
        self.arg = arg
        self.arg_value = arg_value

class SelectorDefaultArgValueNode(CmdNode):
    """
    Attributes:
        arg_value (Token, NumberRangeNode, IntRangeNode)
        negated (bool)
    """
    def __init__(self, arg_value, negated=False):
        self.arg_value = arg_value
        self.negated = negated

class SelectorDefaultGroupArgValueNode(CmdNode):
    """
    Attributes:
        arg_values (list of SelectorDefaultArgValueNode objects)
    """
    def __init__(self, arg_values):
        assert_list_types(arg_values, SelectorDefaultArgValueNode)
        self.arg_values = arg_values

class SelectorTagArgsNode(CmdNode):
    """
    Attributes:
        tag_args (list of SelectorTagArgNode objects)
    """
    def __init__(self, tag_args):
        assert_list_types(tag_args, SelectorTagArgNode, duplicate_key=lambda x: x.tag.value)
        self.tag_args = tag_args

class SelectorTagArgNode(CmdNode):
    """
    Attributes:
        tag (Token)
        negated (bool)
    """
    def __init__(self, tag, negated=False):
        assert_type(tag, Token)
        assert_type(negated, bool)
        self.tag = tag
        self.negated = negated

class SelectorNbtArgsNode(CmdNode):
    """
    Attributes:
        nbt_args (list of SelectorTagArgNode objects)
    """
    def __init__(self, nbt_args):
        assert_list_types(nbt_args, SelectorTagArgNode)
        self.nbt_args = nbt_args

class SelectorNbtArgNode(CmdNode):
    """
    Attributes:
        nbt (NbtObjectNode)
        negated (bool)
    """
    def __init__(self, nbt, negated=False):
        assert_type(nbt, NbtObjectNode)
        assert_type(negated, bool)
        self.nbt = nbt
        self.negated = negated

class SelectorAdvancementGroupArgNode(CmdNode):
    """
    Attributes:
        advancements (?)
    """
    def __init__(self, advancements):
        self.advancements = advancements


class NbtNode(CmdNode, ABC):
    pass

class NbtObjectNode(NbtNode):
    """
    Attributes:
        mappings (list of NbtMapNode objects)
    """
    def __init__(self, mappings):
        assert_list_types(mappings, NbtMapNode, duplicate_key=lambda x: x.arg.value)
        self.mappings = mappings

class NbtMapNode(NbtNode):
    """
    Attributes:
        arg (Token)
        value (Token, NbtObjectNode, NbtArrayNode, NbtIntegerNode, NbtFloatNode)
    """
    def __init__(self, arg, value):
        assert_type(arg, Token)
        assert_type(value, Token, NbtObjectNode, NbtArrayNode, NbtIntegerNode, NbtFloatNode)
        self.arg = arg
        self.value = value

class NbtArrayNode(NbtNode):
    """
    Attributes:
        values (list of NbtNode objects)
        type_specifier (Token or None)
    """
    def __init__(self, values, type_specifier=None):
        assert_list_types(values, Token, NbtObjectNode, NbtArrayNode, NbtIntegerNode, NbtFloatNode)
        assert_type(type_specifier, Token, optional=True)
        self.values = values
        self.type_specifier = type_specifier

class NbtIntegerNode(NbtNode):
    """
    Attributes:
        int_value (Token)
        int_type (Token or None)
    """
    def __init__(self, int_value, int_type=None):
        assert_type(int_value, Token)
        assert_type(int_type, Token, optional=True)
        self.int_value = int_value
        self.int_type = int_type

class NbtFloatNode(NbtNode):
    """
    Attributes:
        float_value (Token)
        float_type (Token or None)
    """
    def __init__(self, float_value, float_type=None):
        assert_type(float_value, Token)
        assert_type(float_type, Token, optional=True)
        self.float_value = float_value
        self.float_type = float_type


class JsonNode(CmdNode, ABC):
    pass

class JsonObjectNode(JsonNode):
    """
    Attributes:
        mappings (list of JsonMapNode objects)
    """
    def __init__(self, mappings):
        assert_list_types(mappings, JsonMapNode, duplicate_key=lambda x: x.arg.value)
        self.mappings = mappings

class JsonMapNode(JsonNode):
    """
    Attributes:
        arg (Token)
        value (Token, JsonArrayNode, JsonObjectNode)
    """
    def __init__(self, arg, value):
        assert_type(arg, Token)
        assert_type(value, Token, JsonArrayNode, JsonObjectNode)
        self.arg = arg
        self.value = value

class JsonArrayNode(JsonNode):
    """
    Attributes:
        values (list of Token, JsonArrayNode, JsonObjectNode)
    """
    def __init__(self, values):
        assert_list_types(values, Token, JsonArrayNode, JsonObjectNode)
        self.values = values


class IntRangeNode(CmdNode):
    """
    Note that a range can be a singular number. If so, left_int is the same as right_int

    Attributes:
        min_int (Token or None)
        max_int (Token or None)
        args (tuple of 2 strs or None): Contains the argument for the min int and the max int (eg. (rm, r))
    """
    def __init__(self, min_int, max_int, args=()):
        assert_type(min_int, Token, optional=True)
        assert_type(max_int, Token, optional=True)
        assert not (min_int is None and max_int is None)
        assert_tuple_types(args, str)
        assert len(args) in (0, 2)

        # checks whether the number actually works as a 32 bit signed int
        if min_int is not None:
            assert (-1<<31) <= int(min_int.value) <= ((1<<31)-1)
        if max_int is not None:
            assert (-1<<31) <= int(max_int.value) <= ((1<<31)-1)

        # checks whether the max int is actually greater than or equal to the min int if they both exist
        if min_int is not None and max_int is not None:
            assert int(max_int.value) >= int(min_int.value)

        self.min_int = min_int
        self.max_int = max_int
        self.args = args

class NumberRangeNode(CmdNode):
    """
    Note that a range can be a singular number. If so, min_number is the same as max_number

    Attributes:
        min_number (Token or None)
        max_number (Token or None)
    """
    def __init__(self, min_number, max_number):
        assert_type(min_number, Token, optional=True)
        assert_type(max_number, Token, optional=True)
        assert not (min_number is None and max_number is None)
        self.min_number = min_number
        self.max_number = max_number

class BlockNode(CmdNode):
    """
    Attributes:
        block (Token)
        states (Token or list of BlockStateNode objects or None)
        nbt (NbtObjectNode or None)
    """
    def __init__(self, block, states, nbt=None):
        assert_type(block, Token)
        assert_list_types(states, BlockStateNode) if isinstance(states, list) else assert_type(states, Token, optional=True)
        assert_type(nbt, NbtObjectNode, optional=True)
        self.block = block
        self.states = states
        self.nbt = nbt

class BlockStateNode(CmdNode):
    """
    Attributes:
        arg (Token)
        value (Token)
    """
    def __init__(self, arg, value):
        assert_type(arg, Token)
        assert_type(value, Token)
        self.arg = arg
        self.value = value

class Vec2Node(CmdNode):
    """
    Attributes:
        coord1 (Token)
        coord2 (Token)
    """
    def __init__(self, coord1, coord2):
        assert_type(coord1, Token)
        assert_type(coord2, Token)
        self.coord1 = coord1
        self.coord2 = coord2

class Vec3Node(CmdNode):
    """
    Attributes:
        coord1 (Token)
        coord2 (Token)
        coord3 (Token)
    """
    def __init__(self, coord1, coord2, coord3):
        assert_type(coord1, Token)
        assert_type(coord2, Token)
        assert_type(coord3, Token)
        self.coord1 = coord1
        self.coord2 = coord2
        self.coord3 = coord3

class NamespaceIdNode(CmdNode):
    """
    Attributes:
        namespace (Token or None)
        id_value (Token)
    """
    def __init__(self, id_value, namespace=None):
        assert_type(id_value, Token)
        assert_type(namespace, Token, optional=True)
        self.id_value = id_value
        self.namespace = namespace

class DataPathNode(CmdNode):
    """
    Attributes:
        path_tokens (list of Token objects)
    """
    def __init__(self, path_tokens):
        assert_list_types(path_tokens, Token)
        self.path_tokens = path_tokens


if __name__ == "__main__":
    # testing random hashes
    class A:
        def __init__(self, x, y):
            self.x = x
            self.y = y

    a = A(1, [2, 3, 4])
    some_dict = {a: 1}
    print(some_dict)




if __name__ == "__main__":
    import sys
    sys.path.append("..")
    del sys

from fenalib.config_data import ConfigData
from fenalib.token_classes import TypedToken
from fenalib.number_utils import is_number

config_data = ConfigData()

def is_coord(string):
    """
    Checks whether a string is a valid coordinate type

    Args:
        string (str)

    Returns:
        bool: Whether the token value can be a coordinate
    """
    if config_data.version == "1.12":
        return string and (string[0] in "~" and (len(string) == 1 or is_number(string[1:]))) or is_number(string)
    return string and (string[0] in "~^" and (len(string) == 1 or is_number(string[1:]))) or is_number(string)

def is_coord_token(token):
    """
    Checks whether a singular token is a coordinate

    Args:
        token (Token)

    Returns:
        bool: Whether the token value can be a coordinate
    """
    return is_coord(token.value)

def get_coord_type(coord):
    """
    This assumes the given coordinate is indeed a coord type

    Args:
        coord (str)

    Returns:
        str ("relative", "local", "absolute"): The coordinate type
    """
    coord_dict = {"^": "local", "~": "relative"}
    return coord_dict.get(coord[0], "absolute")

def get_token_coord_type(token):
    return get_coord_type(token.value)

def are_coords(*tokens):
    """
    Checks whether a group of tokens:
        - Have a length of 2 or 3 tokens
        - Have the same type (local, or world (absolute and relative) coordinates)
            - if the length is 2, then it can only contain world coordinates
    """
    # records the supposed global coord type as either "relative" or "local", as to skip absolute

    overall_coord_type = None
    if not (2 <= len(tokens) <= 3):
        return False

    for token in tokens:
        if not is_coord_token(token):
            return False

        coord_type = get_token_coord_type(token)
        if coord_type in ("relative", "absolute"):
            coord_type = "world"

        if overall_coord_type is None:
            overall_coord_type = coord_type
        elif overall_coord_type != coord_type:
            return False

    if overall_coord_type == "local" and len(tokens) == 2:
        return False
    return True

if __name__ == "__main__":
    # test cases for this are under test_fena/ now
    pass


"""
Simple assert utils to make sure that an object matches some specified type
or the items in an iterable match some specified type

The main difference between using this function and using assert isinstance(...)
is that it provides a more in-depth error message
"""

def assert_type(checked_object, *valid_types, additional_message=None, optional=False):
    """
    Checks whether an object is an instance of any of the valid types

    Args:
        checked_object (any)
        valid_types (any)
        additional_message (str or None)
        optional (bool): Whether the object can be a None object or not

    Examples:
        >>> assert_type([], list, int)
        >>> assert_type(45, list, int)
        >>> assert_type("yep", list, int)
        Traceback (most recent call last):
            ...
        AssertionError: Expected type of 'yep' to be one of (<class 'list'>, <class 'int'>) but got <class 'str'>

        >>> assert_type((25,), list)
        Traceback (most recent call last):
            ...
        AssertionError: Expected type of (25,) to be <class 'list'> but got <class 'tuple'>
    """
    assert isinstance(additional_message, str) or additional_message is None
    assert isinstance(optional, bool)

    if not (isinstance(checked_object, valid_types) or (optional and checked_object is None)):
        if len(valid_types) == 1:
            checked_types_str = f"{valid_types[0]}"
        else:
            checked_types_str = f"one of {valid_types}"

        if optional:
            checked_types_str += " or None"

        full_message = f"Expected type of {checked_object!r} to be {checked_types_str} but got {type(checked_object)}"
        #full_message = f"Expected type of {type(checked_object)} to be {checked_types_str} but got {type(checked_object)}"
        if additional_message is not None:
            full_message += f" {additional_message}"
        raise AssertionError(full_message)

def assert_iterable_types(checked_iterable, *valid_types, duplicate_key=None):
    """
    Checks whether all objects in a iterable are instances of one of the valid types

    Args:
        checked_iterable (iterable)
        valid_types (any)
        duplicate_key (None or callable): A function to give the value to check for duplicates
            Note that if you want to detect duplicates with itself, just do `duplicate_key=lambda x: x`
            If the option is None, no duplicate checking will be done

    Examples:
        >>> assert_iterable_types([1, 3, 2], int)
        >>> assert_iterable_types(["a", "b", "c"], str)
        >>> assert_iterable_types(("a", "b", "c"), int, str)
        >>> assert_iterable_types(["a", "b", "c"], int)
        Traceback (most recent call last):
            ...
        AssertionError: Expected type of 'a' to be <class 'int'> but got <class 'str'> in index 0 of iterable ['a', 'b', 'c']

        >>> assert_iterable_types([23, "b", 327.4], int, str)
        Traceback (most recent call last):
            ...
        AssertionError: Expected type of 327.4 to be one of (<class 'int'>, <class 'str'>) but got <class 'float'> in index 2 of iterable [23, 'b', 327.4]

        >>> assert_iterable_types([23, "b", 327.4], int, str, duplicate_key=lambda x: x)
        Traceback (most recent call last):
            ...
        AssertionError: Expected type of 327.4 to be one of (<class 'int'>, <class 'str'>) but got <class 'float'> in index 2 of iterable [23, 'b', 327.4]

        >>> assert_iterable_types(("a", "b", "c"), str, duplicate_key=lambda x: x)
        >>> assert_iterable_types(("a", "b", "a"), str, duplicate_key=lambda x: x)
        Traceback (most recent call last):
            ...
        AssertionError: Found a duplicate of 'a' in 'a'

        >>> assert_iterable_types(["123", "142", "225"], str, duplicate_key=lambda x: x[0])
        Traceback (most recent call last):
            ...
        AssertionError: Found a duplicate of '1' in '142'
    """
    for index, item in enumerate(checked_iterable):
        assert_type(item, *valid_types, additional_message=f"in index {index} of iterable {checked_iterable}")

    # checks for duplicates
    if duplicate_key is not None:
        assert callable(duplicate_key), "The duplicate key must be a function"

        # uses a set because only the containment of an item matters
        found_objects = set()

        # iterates through each checked object to see if its within the found objects
        for checked_object in checked_iterable:
            result_object = duplicate_key(checked_object)
            # assert result_object not in found_objects, f"Found a duplicate of {result_object!r} in iterable {checked_iterable}"
            assert result_object not in found_objects, f"Found a duplicate of {result_object!r} in {result_object}"
            found_objects.add(result_object)

def assert_list_types(checked_list, *valid_types, **kwargs):
    assert_type(checked_list, list)
    assert_iterable_types(checked_list, *valid_types, **kwargs)

def assert_tuple_types(checked_tuple, *valid_types, **kwargs):
    assert_type(checked_tuple, tuple)
    assert_iterable_types(checked_tuple, *valid_types, **kwargs)

if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
implement the parser for pyexpander.

NOTE: THIS IS A MODIFIED VERSION OF PYEXPANDER as of July 2018
License: https://bitbucket.org/goetzpf/pyexpander/src/b466de6fd801545650edfa790a18f022dc7e151a/LICENSE?at=default&fileviewer=file-view-default
Original: http://pyexpander.sourceforge.net/
"""
import bisect
import re

__version__= "1.8.3" #VERSION#

# pylint: disable= invalid-name

# we use '\n' as line separator and rely on python's built in line end
# conversion to '\n' on all platforms.
# You may howver, use change_linesep() to change the line separator.

LINESEP= "\n"
LINESEP_LEN= len(LINESEP)

def change_linesep(sep):
    """change line separator, this is here just for tests.
    """
    # pylint: disable= global-statement
    global LINESEP, LINESEP_LEN
    LINESEP= sep
    LINESEP_LEN= len(sep)

class IndexedString(object):
    """a string together with row column information.

    Here is an example:

    >>> txt='''01234
    ... 67
    ... 9abcd'''
    >>> l=IndexedString(txt)
    >>> l.rowcol(0)
    (1, 1)
    >>> l.rowcol(1)
    (1, 2)
    >>> l.rowcol(4)
    (1, 5)
    >>> l.rowcol(5)
    (1, 6)
    >>> l.rowcol(6)
    (2, 1)
    >>> l.rowcol(7)
    (2, 2)
    >>> l.rowcol(8)
    (2, 3)
    >>> l.rowcol(9)
    (3, 1)
    >>> l.rowcol(13)
    (3, 5)
    >>> l.rowcol(14)
    (3, 6)
    >>> l.rowcol(16)
    (3, 8)
    """
    # pylint: disable= too-few-public-methods
    def __init__(self, st):
        self._st=st
        self._lines=None
        self._positions=None
    def _list(self):
        """calculate and remember positions where lines begin."""
        l= len(self._st)
        pos=0
        self._lines=[1]
        self._positions=[0]
        lineno=1
        while True:
            # look for the standard lineseparator in the string:
            p= self._st.find(LINESEP, pos)
            if p<0:
                # not found
                break
            pos= p+LINESEP_LEN
            if pos>=l:
                break
            lineno+=1
            self._lines.append(lineno)
            self._positions.append(pos)
    def rowcol(self,pos):
        """calculate (row,column) from a string position."""
        if self._lines is None:
            self._list()
        idx= bisect.bisect_right(self._positions, pos)-1
        off= self._positions[idx]
        return(self._lines[idx], pos-off+1)
    def st(self):
        """return the raw string."""
        return self._st
    def __str__(self):
        return "IndexedString(...)"
    def __repr__(self):
        # Note: if repr(some object) gets too long since
        # repr(IndexedString(..)) basically prints the whole input file
        # you may in-comment the following line in order to make
        # the output shorter:
        #return "IndexedString(...)"
        return "IndexedString(%s)" % repr(self._st)

class ParseException(Exception):
    """used for Exceptions in this module."""
    def __init__(self, value, pos=None, rowcol=None):
        super(ParseException, self).__init__(value, pos, rowcol)
        self.value = value
        self.pos= pos
        self.rowcol= rowcol
    def __str__(self):
        if self.rowcol is not None:
            return "%s line %d, col %d" % \
                   (self.value,self.rowcol[0],self.rowcol[1])
        elif self.pos is not None:
            return "%s position: %d" % (self.value,self.pos)
        return "%s" % self.value

rx_pyIdent= re.compile(r'([A-Za-z_][\w\.]*)$')

rx_csv=re.compile(r'\s*,\s*')

def scanPyIdentList(st):
    """scan a list of python identifiers.

    Here are some examples:

    >>> scanPyIdentList("a,b")
    ['a', 'b']
    >>> scanPyIdentList("a,b.d, c")
    ['a', 'b.d', 'c']
    >>> scanPyIdentList("a,b.d, c&")
    Traceback (most recent call last):
        ...
    ParseException: list of python identifiers expected
    """
    lst= re.split(rx_csv, st)
    for elm in lst:
        m= rx_pyIdent.match(elm)
        if m is None:
            raise ParseException("list of python identifiers expected")
    return lst

rx_py_in= re.compile(r'^\s*(.*?)\s*\b(in)\b\s*(.*?)\s*$')
def scanPyIn(st):
    """scan a python "in" statement.

    Here are some examples:

    >>> scanPyIn(" (a,b) in k.items() ")
    ('(a,b)', 'in', 'k.items()')
    """
    m= rx_py_in.match(st)
    if m is None:
        raise ParseException("python \"in\" expression expected")
    return m.groups()

rx_bracketed= re.compile(r'\{[A-Za-z_]\w*\}')

def parseBracketed(idxst,pos):
    """parse an identifier in curly brackets.

    Here are some examples:

    >>> def test(st,pos):
    ...     idxst= IndexedString(st)
    ...     (a,b)= parseBracketed(idxst,pos)
    ...     print(st[a:b])
    ...
    >>> test(r'{abc}',0)
    {abc}
    >>> test(r'{ab8c}',0)
    {ab8c}
    >>> test(r'{c}',0)
    {c}
    >>> test(r'{}',0)
    Traceback (most recent call last):
        ...
    ParseException: command enclosed in curly brackets at line 1, col 1
    >>> test(r'{abc',0)
    Traceback (most recent call last):
        ...
    ParseException: command enclosed in curly brackets at line 1, col 1
    >>> test(r'x{ab8c}',1)
    {ab8c}
    """
    if not isinstance(idxst, IndexedString):
        raise TypeError("idxst par wrong: %s" % repr(idxst))
    st= idxst.st()
    m= rx_bracketed.match(st,pos)
    if m is None:
        raise ParseException("command enclosed in curly brackets at",
                             rowcol= idxst.rowcol(pos))
    return(pos,m.end())

rx_StringLiteralStart= re.compile(r'''(UR|Ur|uR|ur|r|u|R|U|)("""|''' + \
                                  """'''""" + \
                                  r'''|'|")''')
def parseStringLiteral(idxst,pos):
    r"""parse a python string literal.

    returns 2 numbers, the index where the string starts and
    the index of the first character *after* the string

    Here are some examples:

    >>> def test(st,pos):
    ...     idxst= IndexedString(st)
    ...     (a,b)= parseStringLiteral(idxst,pos)
    ...     print(st[a:b])
    ...

    >>> test(r'''"abc"''',0)
    "abc"

    >>> test("'''ab'c'd'''",0)
    '''ab'c'd'''
    >>> test("'''ab'cd''''",0)
    '''ab'cd'''

    >>> test(r'''U"abc"''',0)
    U"abc"
    >>> test(r'''xU"abc"''',1)
    U"abc"
    >>> test(r'''xUr"abc"''',1)
    Ur"abc"

    >>> test(r'''xUr"ab\\"c"''',1)
    Ur"ab\\"

    >>> test(r'''xUr"ab\"c"''',1)
    Ur"ab\"c"
    >>> test(r'''xUr"ab\"c"''',0)
    Traceback (most recent call last):
        ...
    ParseException: start of string expected at line 1, col 1
    >>> test(r'''"ab''',0)
    Traceback (most recent call last):
        ...
    ParseException: end of string not found at line 1, col 1
    >>> test(r"'''ab'",0)
    Traceback (most recent call last):
        ...
    ParseException: end of string not found at line 1, col 1
    >>> test(r'''"ab\"''',0)
    Traceback (most recent call last):
        ...
    ParseException: end of string not found at line 1, col 1
    """
    if not isinstance(idxst, IndexedString):
        raise TypeError("idxst par wrong: %s" % repr(idxst))
    st= idxst.st()
    m= rx_StringLiteralStart.match(st,pos)
    if m is None:
        raise ParseException("start of string expected at",
                             rowcol= idxst.rowcol(pos))
    prefix= m.group(1)
    starter= m.group(2) # """ or ''' or " or '
    #is_unicode= False
    #is_raw= False
    #if -1!=prefix.find("r"):
    #    is_raw= True
    #elif -1!=prefix.find("R"):
    #    is_raw= True
    #if -1!=prefix.find("u"):
    #    is_unicode= True
    #elif -1!=prefix.find("U"):
    #    is_unicode= True
    startpos= pos+len(prefix)+len(starter)
    while True:
        idx= st.find(starter, startpos)
        # if startpos>len(st), idx is also -1
        if idx==-1:
            raise ParseException("end of string not found at",
                                 rowcol= idxst.rowcol(pos))
        if st[idx-1]=="\\":
            # maybe escaped quote char
            e= None
            try:
                if st[idx-2]!="\\":
                    # only then it is an escaped quote char
                    startpos= idx+1
                    continue
            except IndexError as _e:
                e= _e
            if e is not None:
                raise ParseException("end of string not found at",
                                     rowcol= idxst.rowcol(pos))
        break
    if len(starter)==1:
        # simple single quoted string
        return(pos,idx+1)
    return(pos,idx+3)

def parseComment(idxst,pos):
    r"""parse a python comment.

    Here are some examples:

    >>> import os
    >>> def test(st,pos,sep=None):
    ...     if sep:
    ...         change_linesep(sep)
    ...     idxst= IndexedString(st)
    ...     (a,b)= parseComment(idxst,pos)
    ...     print(repr(st[a:b]))
    ...     change_linesep(os.linesep)
    ...
    >>> test("#abc",0)
    '#abc'
    >>> test("#abc\nef",0,"\n")
    '#abc\n'
    >>> test("#abc\r\nef",0,"\r\n")
    '#abc\r\n'
    >>> test("xy#abc",2)
    '#abc'
    >>> test("xy#abc\nef",2,"\n")
    '#abc\n'
    >>> test("xy#abc\nef",3)
    Traceback (most recent call last):
        ...
    ParseException: start of comment not found at line 1, col 4
    """
    if not isinstance(idxst, IndexedString):
        raise TypeError("idxst par wrong: %s" % repr(idxst))
    st= idxst.st()
    if st[pos]!="#":
        raise ParseException("start of comment not found at",
                             rowcol= idxst.rowcol(pos))
    idx_lf= st.find(LINESEP,pos+1)
    if idx_lf==-1:
        return(pos, len(st))
    return(pos,idx_lf+LINESEP_LEN)

rx_CodePart= re.compile(r'''((?:UR|Ur|uR|ur|r|u|R|U|)(?:"""|''' + """'''""" + \
                        r'''|'|")|#|\(|\))''')
def parseCode(idxst,pos):
    r"""parse python code, it MUST start with a '('.

    Here are some examples:

    >>> def test(st,pos):
    ...     idxst= IndexedString(st)
    ...     (a,b)= parseCode(idxst,pos)
    ...     print(st[a:b])
    ...
    >>> test(r'(a+b)',0)
    (a+b)
    >>> test(r'(a+(b*c))',0)
    (a+(b*c))
    >>> test(r'(a+(b*c)+")")',0)
    (a+(b*c)+")")
    >>> test(r"(a+(b*c)+''')''')",0)
    (a+(b*c)+''')''')
    >>> test(r"(a+(b*c)+''')'''+# comment )\n)",0)
    Traceback (most recent call last):
        ...
    ParseException: end of bracket expression not found at line 1, col 1
    >>>
    >>> test("(a+(b*c)+''')'''+# comment )\n)",0)
    (a+(b*c)+''')'''+# comment )
    )
    """
    if not isinstance(idxst, IndexedString):
        raise TypeError("idxst par wrong: %s" % repr(idxst))
    st= idxst.st()
    if st[pos]!="(":
        raise ParseException("start of bracket expression not found at",
                             rowcol= idxst.rowcol(pos))
    startpos= pos+1
    while True:
        m= rx_CodePart.search(st, startpos)
        if m is None:
            raise ParseException("end of bracket expression not found at",
                                 rowcol= idxst.rowcol(pos))
        matched= m.group(1)
        if matched=="#":
            # a comment
            (_,b)= parseComment(idxst, m.start())
            startpos= b
            continue
        if matched=="(":
            # an inner bracket
            (_,b)= parseCode(idxst, m.start())
            startpos= b
            continue
        if matched==")":
            return(pos,m.start()+1)
        # from here it must be a string literal
        (_,b)= parseStringLiteral(idxst, m.start())
        startpos= b
        continue

class ParsedItem(object):
    """base class of parsed items."""
    def __init__(self, idxst, start, end):
        if not isinstance(idxst, IndexedString):
            raise TypeError("idxst par wrong: %s" % repr(idxst))
        self._idxst= idxst
        self._start= start
        self._end= end
    def string(self):
        """return the string that represents the ParsedItem."""
        return self._idxst.st()[self._start:self._end+1]
    def start(self):
        """return the start of the ParsedItem in the source string."""
        return self._start
    def end(self):
        """return the end of the ParsedItem in the source string."""
        return self._end
    def rowcol(self, pos= None):
        """calculate (row,column) from a string position."""
        if pos is None:
            pos= self.start()
        return self._idxst.rowcol(pos)
    def positions(self):
        """return start and end of ParsedItem in the source string."""
        return "(%d, %d)" % (self._start, self._end)
    def __str__(self):
        return "('%s', %s, %s)" % (self.__class__.__name__, \
               self.positions(), repr(self.string()))
    def __repr__(self):
        return "%s(%s, %s, %s)" % (self.__class__.__name__, \
                repr(self._idxst), repr(self._start), repr(self._end))

class ParsedLiteral(ParsedItem):
    """class of a parsed literal.

    A literal is a substring in the input that shouldn't be modified by
    pyexpander.
    """
    def __init__(self, idxst, start, end):
        ParsedItem.__init__(self, idxst, start, end)

class ParsedComment(ParsedItem):
    """class of a parsed comment.

    A comment in pyexpander starts with '$#'.
    """
    def __init__(self, idxst, start, end):
        ParsedItem.__init__(self, idxst, start, end)

class ParsedVar(ParsedItem):
    """class of a parsed variable.

    A variable in pyexpander has the form "$(identifier)".
    """
    def __init__(self, idxst, start, end):
        ParsedItem.__init__(self, idxst, start, end)

class ParsedEval(ParsedItem):
    """class of an pyexpander expression.

    A pyexpander expression has the form "$(expression)" e.g. "$(a+1)". This is
    different from ParsedVar where the string within the brackets is a simple
    identifier.
    """
    def __init__(self, idxst, start, end):
        ParsedItem.__init__(self, idxst, start, end)

class ParsedPureCommand(ParsedItem):
    """class of a pyexpander command without arguments.

    A pure command has the form "$name". Such a command has no arguments which
    would be enclosed in round brackets immediately following the name.
    """
    def __init__(self, idxst, start, end):
        ParsedItem.__init__(self, idxst, start, end)

class ParsedCommand(ParsedItem):
    """class of a pyexpander command with arguments.

    A command has the form "$name(argument1, argument2, ...)".
    """
    def __init__(self, idxst, start, end, ident):
        ParsedItem.__init__(self, idxst, start, end)
        self.ident= ident
    def args(self):
        """return the arguments of the command."""
        return self.string()
    def __str__(self):
        return "('%s', %s, %s, %s)" % (self.__class__.__name__, \
               self.positions(), repr(self.string()), \
               repr(self.ident))
    def __repr__(self):
        return "%s(%s, %s, %s, %s)" % (self.__class__.__name__, \
                repr(self._idxst), repr(self._start), repr(self._end), \
                repr(self.ident))


rx_DollarFollows= re.compile(r'([A-Za-z_]\w*|\(|\{|#)')

def parseDollar(idxst, pos):
    r"""parse things that follow a dollar.

    Here are some examples:

    >>> def test(st,pos):
    ...   idxst= IndexedString(st)
    ...   (p,elm)= parseDollar(idxst,pos)
    ...   print("Parsed: %s" % elm)
    ...   print("rest of string:%s" % st[p:])
    ...
    >>> test("$abc",0)
    Parsed: ('ParsedPureCommand', (1, 3), 'abc')
    rest of string:
    >>> test("$abc%&/",0)
    Parsed: ('ParsedPureCommand', (1, 3), 'abc')
    rest of string:%&/
    >>> test("$abc(2*3)",0)
    Parsed: ('ParsedCommand', (5, 7), '2*3', 'abc')
    rest of string:
    >>> test(" $abc(2*sin(x))",1)
    Parsed: ('ParsedCommand', (6, 13), '2*sin(x)', 'abc')
    rest of string:
    >>> test(" $abc(2*sin(x))bn",1)
    Parsed: ('ParsedCommand', (6, 13), '2*sin(x)', 'abc')
    rest of string:bn
    >>> test(" $# a comment\nnew line",1)
    Parsed: ('ParsedComment', (3, 13), ' a comment\n')
    rest of string:new line
    >>> test("$(abc)",0)
    Parsed: ('ParsedVar', (2, 4), 'abc')
    rest of string:
    >>> test("$(abc*2)",0)
    Parsed: ('ParsedEval', (2, 6), 'abc*2')
    rest of string:
    >>> test(" $(2*x(y))abc",1)
    Parsed: ('ParsedEval', (3, 8), '2*x(y)')
    rest of string:abc
    """
    if not isinstance(idxst, IndexedString):
        raise TypeError("idxst par wrong: %s" % repr(idxst))
    st= idxst.st()
    if st[pos]!="$":
        raise ParseException("'$' expected at",
                             rowcol= idxst.rowcol(pos))
    m= rx_DollarFollows.match(st, pos+1)
    if m is None:
        raise ParseException("unexpected characters after '$' at",
                             rowcol= idxst.rowcol(pos))
    matched= m.group(1)
    if matched=="#":
        # an expander comment
        (a, b)= parseComment(idxst, pos+1)
        elm= ParsedComment(idxst, a+1, b-1)
        return (b, elm)
    if matched=="(":
        # pylint: disable= redefined-variable-type
        (a, b)= parseCode(idxst, pos+1)
        m_ident= rx_pyIdent.match(st, a+1, b-1)
        if m_ident is not None:
            elm= ParsedVar(idxst, a+1, b-2)
        else:
            elm= ParsedEval(idxst, a+1, b-2)
        return (b, elm)
    if matched=="{":
        # a purecommand enclosed in "{}" brackets
        (a, b)= parseBracketed(idxst, pos+1)
        elm= ParsedPureCommand(idxst, a+1, b-2)
        return (b, elm)
    # from here: a purecommand or a command
    try:
        nextchar= st[m.end()]
    except IndexError as _:
        nextchar= None
    if nextchar=="(":
        (a, b)= parseCode(idxst, m.end())
        elm= ParsedCommand(idxst, a+1, b-2, matched)
        return (b, elm)
    elm= ParsedPureCommand(idxst, pos+1, m.end()-1)
    return (m.end(), elm)

def parseBackslash(idxst, pos):
    r"""parses a backslash.

    >>> import os
    >>> def test(st,pos,sep=None):
    ...     if sep:
    ...         change_linesep(sep)
    ...     idxst= IndexedString(st)
    ...     (p,elm)= parseBackslash(idxst,pos)
    ...     print("Parsed: %s" % elm)
    ...     print("rest of string:%s" % repr(st[p:]))
    ...     change_linesep(os.linesep)
    ...
    >>> test(r"\abc",0)
    Parsed: ('ParsedLiteral', (0, 0), '\\')
    rest of string:'abc'
    >>> test("\\",0)
    Parsed: ('ParsedLiteral', (0, 0), '\\')
    rest of string:''
    >>> test("\\\rab",0,"\r")
    Parsed: None
    rest of string:'ab'
    >>> test("\\\rab",0,"\n")
    Parsed: ('ParsedLiteral', (0, 0), '\\')
    rest of string:'\rab'
    >>> test("\\\rab",0,"\r\n")
    Parsed: ('ParsedLiteral', (0, 0), '\\')
    rest of string:'\rab'
    >>> test("\\\nab",0,"\r")
    Parsed: ('ParsedLiteral', (0, 0), '\\')
    rest of string:'\nab'
    >>> test("\\\nab",0,"\n")
    Parsed: None
    rest of string:'ab'
    >>> test("\\\nab",0,"\r\n")
    Parsed: ('ParsedLiteral', (0, 0), '\\')
    rest of string:'\nab'
    >>> test("\\\r\nab",0,"\r")
    Parsed: None
    rest of string:'\nab'
    >>> test("\\\r\nab",0,"\n")
    Parsed: ('ParsedLiteral', (0, 0), '\\')
    rest of string:'\r\nab'
    >>> test("\\\r\nab",0,"\r\n")
    Parsed: None
    rest of string:'ab'
    """
    # backslash found
    if not isinstance(idxst, IndexedString):
        raise TypeError("idxst par wrong: %s" % repr(idxst))
    st= idxst.st()
    if st[pos]!="\\":
        raise ParseException("backslash expected at",
                             rowcol= idxst.rowcol(pos))
    l_= len(st)
    if pos+1>=l_:
        # no more characters
        elm= ParsedLiteral(idxst, pos, pos)
        return (pos+1, elm)
    else:
        # at least one more character
        nextchar= st[pos+1]
        if nextchar=="\\":
            elm= ParsedLiteral(idxst, pos, pos)
            return (pos+2, elm)
        if nextchar=="$":
            elm= ParsedLiteral(idxst, pos+1, pos+1)
            return (pos+2, elm)
        if pos+LINESEP_LEN>=l_:
            # not enough characters for LINESEP
            elm= ParsedLiteral(idxst, pos, pos)
            return (pos+1, elm)
        else:
            for i in range(LINESEP_LEN):
                if st[pos+1+i]!=LINESEP[i]:
                    # no line separator follows
                    elm= ParsedLiteral(idxst, pos, pos)
                    return (pos+1, elm)
            return(pos+LINESEP_LEN+1, None)

rx_top= re.compile(r'(\$|\\)')
def parseAll(idxst, pos):
    r"""parse everything.

    >>> def test(st,pos):
    ...     idxst= IndexedString(st)
    ...     pprint(parseAll(idxst,pos))
    ...

    >>> test("abc",0)
    ('ParsedLiteral', (0, 2), 'abc')
    >>> test("abc$xyz",0)
    ('ParsedLiteral', (0, 2), 'abc')
    ('ParsedPureCommand', (4, 6), 'xyz')
    >>> test("abc${xyz}efg",0)
    ('ParsedLiteral', (0, 2), 'abc')
    ('ParsedPureCommand', (5, 7), 'xyz')
    ('ParsedLiteral', (9, 11), 'efg')
    >>> test("abc$xyz(2*4)",0)
    ('ParsedLiteral', (0, 2), 'abc')
    ('ParsedCommand', (8, 10), '2*4', 'xyz')
    >>> test("abc$(2*4)ab",0)
    ('ParsedLiteral', (0, 2), 'abc')
    ('ParsedEval', (5, 7), '2*4')
    ('ParsedLiteral', (9, 10), 'ab')
    >>> test("abc\\$(2*4)ab",0)
    ('ParsedLiteral', (0, 2), 'abc')
    ('ParsedLiteral', (4, 4), '$')
    ('ParsedLiteral', (5, 11), '(2*4)ab')
    >>> test("ab$func(1+2)\\\nnew line",0)
    ('ParsedLiteral', (0, 1), 'ab')
    ('ParsedCommand', (8, 10), '1+2', 'func')
    ('ParsedLiteral', (14, 21), 'new line')
    >>> test("ab$func(1+2)\nnew line",0)
    ('ParsedLiteral', (0, 1), 'ab')
    ('ParsedCommand', (8, 10), '1+2', 'func')
    ('ParsedLiteral', (12, 20), '\nnew line')
    >>> test("ab$(xyz)(56)",0)
    ('ParsedLiteral', (0, 1), 'ab')
    ('ParsedVar', (4, 6), 'xyz')
    ('ParsedLiteral', (8, 11), '(56)')

    >>> test(r'''
    ... Some text with a macro: $(xy)
    ... an escaped dollar: \$(xy)
    ... a macro within letters: abc${xy}def
    ... a pyexpander command structure:
    ... $if(a=1)
    ... here
    ... $else
    ... there
    ... $endif
    ... now a continued\
    ... line
    ... from here:$# the rest is a comment
    ... now an escaped continued\\
    ... line
    ... ''',0)
    ('ParsedLiteral', (0, 24), '\nSome text with a macro: ')
    ('ParsedVar', (27, 28), 'xy')
    ('ParsedLiteral', (30, 49), '\nan escaped dollar: ')
    ('ParsedLiteral', (51, 51), '$')
    ('ParsedLiteral', (52, 83), '(xy)\na macro within letters: abc')
    ('ParsedPureCommand', (86, 87), 'xy')
    ('ParsedLiteral', (89, 124), 'def\na pyexpander command structure:\n')
    ('ParsedCommand', (129, 131), 'a=1', 'if')
    ('ParsedLiteral', (133, 138), '\nhere\n')
    ('ParsedPureCommand', (140, 143), 'else')
    ('ParsedLiteral', (144, 150), '\nthere\n')
    ('ParsedPureCommand', (152, 156), 'endif')
    ('ParsedLiteral', (157, 172), '\nnow a continued')
    ('ParsedLiteral', (175, 189), 'line\nfrom here:')
    ('ParsedComment', (192, 214), ' the rest is a comment\n')
    ('ParsedLiteral', (215, 238), 'now an escaped continued')
    ('ParsedLiteral', (239, 239), '\\')
    ('ParsedLiteral', (241, 246), '\nline\n')
    """
    if not isinstance(idxst, IndexedString):
        raise TypeError("idxst par wrong: %s" % repr(idxst))
    st= idxst.st()
    parselist=[]
    l= len(st)
    while True:
        if pos>=l:
            return parselist
        m= rx_top.search(st, pos)
        if m is None:
            parselist.append(ParsedLiteral(idxst, pos, len(st)-1))
            return parselist
        if m.start()>pos:
            parselist.append(ParsedLiteral(idxst, pos, m.start()-1))
        if m.group(1)=="\\":
            (p, elm)= parseBackslash(idxst, m.start())
            if elm is not None:
                parselist.append(elm)
            pos= p
            continue
        # from here it must be a dollar sign
        (pos, elm)= parseDollar(idxst, m.start())
        parselist.append(elm)
        continue

def pprint(parselist):
    """pretty print a parselist."""
    for elm in parselist:
        print(str(elm))

def _test():
    """perform the doctest tests."""
    import doctest
    print("testing...")
    doctest.testmod()
    print("done")

if __name__ == "__main__":
    _test()



"""
The main pyexpander library.

NOTE: THIS IS A MODIFIED VERSION OF PYEXPANDER as of July 2018
License: https://bitbucket.org/goetzpf/pyexpander/src/b466de6fd801545650edfa790a18f022dc7e151a/LICENSE?at=default&fileviewer=file-view-default
Original: http://pyexpander.sourceforge.net/
"""
# pylint: disable=too-many-lines

import re
import os
import os.path
import inspect

import sys
import keyword

if __name__ == "__main__":
    # if this module is directly called like a script, we have to add the path
    # ".." to the python search path in order to find modules named
    # "pyexpander.[module]".
    sys.path.append("..")

# pylint: disable=wrong-import-position
# pylint: disable=invalid-name

import fena_pyexpander.parser as EP

__version__= "1.8.3" #VERSION#
assert __version__==EP.__version__

# ---------------------------------------------
# constants
# ---------------------------------------------

PY_KEYWORDS= set(keyword.kwlist)

# length of line separator
LINESEP_LEN= len(os.linesep)

PURE_CMD_KEYWORDS= set([ \
                        "else",
                        "endif",
                        "endfor",
                        "endwhile",
                        "endmacro",
                        "begin",
                        "end",
                       ])

CMD_KEYWORDS=      set([ \
                        "py",
                        "include",
                        "include_begin",
                        "template",
                        "subst",
                        "pattern",
                        "default",
                        "if",
                        "elif",
                        "for",
                        "for_begin",
                        "while",
                        "while_begin",
                        "macro",
                        "nonlocal",
                        "extend",
                        "extend_expr",
                       ])

KEYWORDS= PURE_CMD_KEYWORDS | CMD_KEYWORDS | PY_KEYWORDS

# ---------------------------------------------
# dump utilities
# ---------------------------------------------

def _set2str(val):
    """convert an iterable to the repr string of a set."""
    elms= sorted(list(val))
    return "set(%s)" % repr(elms)

def _pr_set(val):
    """print an iterable as the repr string of a set."""
    print(_set2str(val))

def find_file(filename, include_paths):
    """find a file in a list of include paths.

    include_paths MUST CONTAIN "" in order to search the
    local directory.
    """
    if include_paths is None:
        return None
    for path in include_paths:
        p= os.path.join(path, filename)
        if os.path.exists(p):
            if os.access(p, os.R_OK):
                return p
            print("warning: file \"%s\" found but it is not readable" % \
                  p)
    return None

# ---------------------------------------------
# helper functions
# ---------------------------------------------

def keyword_check(identifiers):
    """indentifiers must not be identical to keywords.

    This function may raise an exception.
    """
    s= set(identifiers).intersection(KEYWORDS)
    if s:
        lst= ", ".join(["'%s'" % e for e in sorted(s)])
        raise ValueError("keywords %s cannot be used as identifiers" % lst)

# ---------------------------------------------
# parse a string or a file
# ---------------------------------------------

def parseString(st):
    """parse a string."""
    return EP.parseAll(EP.IndexedString(st), 0)

def parseFile(filename, no_stdin_warning):
    """parse a file."""
    if filename is None:
        if not no_stdin_warning:
            sys.stderr.write("(reading from stdin)\n")
        try:
            st= sys.stdin.read()
        except KeyboardInterrupt:
            sys.exit(" interrupted\n")
    else:
        with open(filename, "rt") as f:
            st= f.read()
    return parseString(st)

# ---------------------------------------------
# Result text class
# ---------------------------------------------

class ResultText(object):
    """basically a list of strings with a current column property.
    """
    def __init__(self):
        """initialize the object."""
        self._list= []
        self._column= -1
    @staticmethod
    def current_column(st):
        r"""find current column if the string is printed.

        Note: With a string ending with '\n' this returns 1.

        Here are some examples: >>> current_column("")
        -1
        >>> ResultText.current_column("ab")
        -1
        >>> ResultText.current_column("\nab")
        3
        >>> ResultText.current_column("\nab\n")
        1
        >>> ResultText.current_column("\nab\na")
        2
        """
        idx= st.rfind(os.linesep)
        if idx<0: # not found
            return -1 # unknown column
        return len(st)-idx
    def append(self, text):
        """append some text."""
        c= self.__class__.current_column(text)
        if c<0:
            if self._column>0:
                self._column+= len(text)
        else:
            self._column= c
        self._list.append(text)
    def list_(self):
        """return internal list."""
        return self._list
    def column(self):
        """return current column."""
        return self._column

# ---------------------------------------------
# process parse-list
# ---------------------------------------------

class Block(object):
    """class that represents a block in the expander language.

    Each block has a parent-pointer, which may be None. Blocks are used to
    represent syntactical blocks in the expander language. They manage the
    global variable dictionary and some more properties that are needed during
    execution of the program.
    """

    # pylint: disable=too-many-instance-attributes
    def posmsg(self, msg=None, pos=None):
        """return a message together with a position.

        This method is mainly used for error messages. We usually want to print
        filename and the line and column number together with a message. This
        method returns the given message string <msg> together with this
        additional information.
        """
        parts=[]
        if msg is not None:
            parts.append(msg)
        if self.filename is not None:
            parts.append("file \"%s\"" % self.filename)
        try:
            # posmsg is often called with pos==None. So we usually have to
            # determine the position ourselves. Usually posmsg is called when
            # an error occurs when processing self.parse_list[self.lst_pos]. So
            # if we call method rowcol(None) with this element (a ParsedItem
            # object) the *start* of the ParsedItem is calculated to a row and
            # a column and returned.
            p_elm= self.parse_list[self.lst_pos]
            parts.append("line %d, col %d" % p_elm.rowcol(pos))
            return " ".join(parts)
        except IndexError as _:
            return " ".join(parts) + " at unknown position"

    def _append(self, lst, name, val=None):
        """append a line with property information to a list.

        This method is used to create an object dump in method _strlist.
        """
        def elm2str(e):
            """convert an item to a "repr" like string."""
            if isinstance(e,set):
                return _set2str(e)
            return str(e)
        if val is not None:
            lst.append("    %-20s: %s" % (name,val))
        else:
            lst.append("    %-20s= %s" % (name,elm2str(getattr(self,name))))

    def _strlist(self):
        """utility for __str__.

        This function can be used by descendent classes in order to implement
        the __str__ method in a simple way.

        Tested by the testcode in __str__.
        """
        lst= []
        self._append(lst, "has parent",self.previous is not None)
        self._append(lst, "filename")
        self._append(lst, "template")
        self._append(lst, "template_path")
        self._append(lst, "has template_parselist_cache", \
                          self.template_parselist_cache is not None)
        self._append(lst, "exported_syms")
        self._append(lst, "direct_vars")
        self._append(lst, "direct_funcs")
        self._append(lst, "macros", repr(sorted(self.macros.keys())))
        self._append(lst, "new_scope")
        self._append(lst, "skip")
        self._append(lst, "indent")
        self._append(lst, "new_parse_list")
        self._append(lst, "lst_pos")
        self._append(lst, "start_pos")
        if self.lst_pos>=0 and self.lst_pos<len(self.parse_list):
            self._append(lst, "current parse elm",
                         self.parse_list[self.lst_pos])
        return lst

    def __init__(self,
                 previous= None,
                 new_scope= False,
                 filename= None,
                 parse_list=None,
                 external_definitions=None):
        """The Block constructor.

        Properties of a Block:
            parse_list    -- the list with parse objects
            lst_pos       -- the position in the parse_list
            start_pos     -- the start position of the block in the parse_list
            previous      -- the previous block
            new_scope     -- True if the block is also a new scope with
                             respect to global variables.
            indent        -- current automatic indent level, the default is 0
            skip          -- if True, skip a block in the text when the code
                             is interpreted
            globals_      -- a dictionary with the current global variables of
                             the interpreter.
            exported_syms -- a list of symbols (names in globals_) that are
                             copied to the globals_ dictionary of the parent
                             block when the pop() method is called.
            filename      -- name of the interpreted file, may be None
            template      -- name of the file that "$subst" would include,
                             usually None
            template_path -- complete path of template
                             usually None
            template_parselist_cache --
                             parselist of the template, usually None.
                             This is a kind of optimization

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> print(b)
        Block{
            has parent          : False
            filename            = None
            template            = None
            template_path       = None
            has template_parselist_cache: False
            exported_syms       = []
            direct_vars         = set([])
            direct_funcs        = set([])
            macros              : []
            new_scope           = False
            skip                = False
            indent              = 0
            new_parse_list      = True
            lst_pos             = -1
            start_pos           = 0
        }
        >>> b= Block(b,True,parse_list=[])
        >>> b.print_block_list()
        Block{
            has parent          : False
            filename            = None
            template            = None
            template_path       = None
            has template_parselist_cache: False
            exported_syms       = []
            direct_vars         = set([])
            direct_funcs        = set([])
            macros              : []
            new_scope           = False
            skip                = False
            indent              = 0
            new_parse_list      = True
            lst_pos             = -1
            start_pos           = 0
        }
        Block{
            has parent          : True
            filename            = None
            template            = None
            template_path       = None
            has template_parselist_cache: False
            exported_syms       = []
            direct_vars         = set([])
            direct_funcs        = set([])
            macros              : []
            new_scope           = True
            skip                = False
            indent              = 0
            new_parse_list      = True
            lst_pos             = -1
            start_pos           = 0
        }
        """

        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        if (filename is None) or (filename==""):
            self.filename= None
        else:
            self.filename= filename
        self.previous= previous
        if previous is None:
            self.new_scope= False
            self.skip= False
            self.indent= 0
            self.globals_= dict()
            self.direct_vars=set()
            self.direct_funcs=set()
            self.macros= dict()
            self.exported_syms=[]
            if parse_list is None:
                raise AssertionError("without previous, parse_list "+\
                                      "is mandatory")
            self.new_parse_list= True
            self.template= None
            self.template_path= None
            self.template_parselist_cache= None
            self.parse_list= parse_list
            self.lst_pos= -1
            self.start_pos= 0
        else:
            if not isinstance(previous,Block):
                raise AssertionError("previous is not a block: %s" % \
                                      str(previous))
            if self.filename is None:
                self.filename= previous.filename
            if parse_list is None:
                self.new_parse_list= False
                self.parse_list= previous.parse_list
                self.lst_pos= previous.lst_pos
                if previous.lst_pos<0:
                    self.start_pos= 0
                else:
                    self.start_pos= previous.lst_pos
            else:
                self.new_parse_list= True
                self.parse_list= parse_list
                self.lst_pos= -1
                self.start_pos= 0
            self.template= previous.template
            self.template_path= previous.template_path
            self.template_parselist_cache= previous.template_parselist_cache
            self.new_scope= new_scope
            self.skip= previous.skip
            self.indent= previous.indent
            if new_scope:
                self.globals_= dict(previous.globals_)
                self.direct_vars= set(previous.direct_vars)
                self.direct_funcs= set(previous.direct_funcs)
                self.macros= dict(previous.macros)
                self.exported_syms=[]
            else:
                self.globals_= previous.globals_
                self.direct_vars= previous.direct_vars
                self.direct_funcs= previous.direct_funcs
                self.macros= previous.macros
                self.exported_syms= previous.exported_syms
        if external_definitions is not None:
            for (k,v) in list(external_definitions.items()):
                self.globals_[k]= v

    def parse_loop(self):
        """loop across the items in the Block.

        returns False if there is nothing more to parse in the current block.
        """
        self.lst_pos+= 1
        return self.lst_pos < len(self.parse_list)

    def parse_elm(self):
        """get the current parse element."""
        return self.parse_list[self.lst_pos]

    def eval_(self, st):
        """perform eval with the global variable dictionary of the block.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> b.exec_("a=2")
        >>> b.eval_("3*a")
        6
        """
        e= None
        try:
            # pylint: disable=eval-used
            return eval(st, self.globals_)
        except Exception as _e:
            # https://docs.python.org/3/library/exceptions.html#BaseException.with_traceback
            tb = sys.exc_info()[2]
            e= _e
        if e is not None:
            raise e.__class__("%s at %s" % (str(e), self.posmsg())).with_traceback(tb)

    def str_eval(self, st):
        """perform eval with the global variable dictionary of the block.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> b.exec_("a=2")
        >>> b.str_eval("3*a")
        '6'
        """
        # pylint: disable=eval-used
        val= self.eval_(st)
        try:
            return str(val)
        except Exception as _:
            sys.stderr.write("error at %s:\n" % self.posmsg())
            raise

    def exec_(self, st):
        """perform exec with the global variable dictionary of the block.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> b.exec_("a=1")
        >>> b["a"]
        1
        >>> b= Block(b,True)
        >>> b["a"]
        1
        >>> b.exec_("a=2")
        >>> b["a"]
        2
        >>> b.previous["a"]
        1
        """

        e= None
        try:
            # pylint: disable=exec-used
            exec(st, self.globals_)
        except Exception as _e:
            tb = sys.exc_info()[2]
            e = _e
        if e is not None:
            # pylint: disable=raising-non-exception
            raise e.__class__("%s at %s" % (str(e), self.posmsg())).with_traceback(tb)

    def __getitem__(self, name):
        """looks up a value in the globals_ dictionary of the block.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> b.globals_["a"]= 5
        >>> b["a"]
        5

        Here we show how symbols with dots '.' are treated. We misuse
        os.path to create a variable os.path.myvar and store a
        reference to the module "os" in the Block "globals_"
        dictionary:

        >>> import os.path
        >>> os.path.myvar=100
        >>> b.globals_["os"]= os
        >>> b["os.path.myvar"]
        100
        """
        e= None
        try:
            # pylint: disable=eval-used
            return eval(name, self.globals_)
        except Exception as _e:
            tb = sys.exc_info()[2]
            e= _e
        if e is not None:
            # pylint: disable=raising-non-exception
            raise e.__class__("%s at %s" % (str(e), self.posmsg())).with_traceback(tb)

    def __setitem__(self, name, val):
        """sets a value in the globals_ dictionary of the block.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> b["a"]= 5
        >>> b.globals_["a"]
        5

        Here we show how symbols with dots '.' are treated. We misuse os.path
        to create a variable os.path.myvar and store a reference to the module
        "os" in the Block "globals_" dictionary:

        >>> import os.path
        >>> b["os"]= os
        >>> b["os.path.myvar"]= 200
        >>> os.path.myvar
        200
        >>> b["os.path.myvar"]
        200
        """
        self.globals_["__pyexpander_buffer"]= val
        e= None
        try:
            # pylint: disable=exec-used
            exec("%s= __pyexpander_buffer" % name, self.globals_)
        except AttributeError as _e:
            e= _e
        if e is not None:
            # pylint: disable=raising-non-exception
            raise e.__class__("%s at %s" % (str(e), self.posmsg()))

    def __delitem__(self, name):
        """deletes a value in the globals_ dictionary of the block.

        Here is an example:

        >>> import pprint
        >>> b= Block(parse_list=[])
        >>> sorted(b.globals_.keys())
        []
        >>> b["a"]= 5
        >>> sorted(b.globals_.keys())
        ['__builtins__', '__pyexpander_buffer', 'a']
        >>> b["a"]
        5
        >>> del b["a"]
        >>> sorted(b.globals_.keys())
        ['__builtins__', '__pyexpander_buffer']
        >>> b["a"]
        Traceback (most recent call last):
            ...
        NameError: name 'a' is not defined at unknown position
        >>> del b["c"]
        Traceback (most recent call last):
            ...
        NameError: name 'c' is not defined at unknown position

        Here we show how symbols with dots '.' are treated. We use
        a dummy class 'T'
        to create a variable main.sub.myvar and store a reference to
        it in the Block's "globals_" dictionary:

        >>> class T(object):
        ...     def __init__(self):
        ...         pass
        ...
        >>> main= T()
        >>> main.sub= T()
        >>> main.sub.myvar=300
        >>> b= Block(parse_list=[])
        >>> b["main"]= main
        >>> b["main.sub.myvar"]
        300
        >>> del b["main.sub.myvar"]
        >>> b["main.sub.myvar"]
        Traceback (most recent call last):
            ...
        AttributeError: 'T' object has no attribute 'myvar'
        >>> main.sub.myvar
        Traceback (most recent call last):
            ...
        AttributeError: 'T' object has no attribute 'myvar'
        """

        e= None
        try:
            # pylint: disable=exec-used
            exec("del %s" % name, self.globals_)
        except NameError as _e:
            e= _e
        if e is not None:
            # pylint: disable=raising-non-exception
            raise e.__class__("%s at %s" % (str(e), self.posmsg()))

    def __len__(self):
        """returns the number of items in the block.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> sorted(b.globals_.keys())
        []
        >>> b["a"]= 5
        >>> b["b"]= 6
        >>> sorted(b.globals_.keys())
        ['__builtins__', '__pyexpander_buffer', 'a', 'b']
        >>> len(b)
        4
        """
        return len(self.globals_)

    def setdefault(self, name, val):
        """return a value from globals, set to a default if it's not defined.

        Here are some examples:

        >>> b= Block(parse_list=[])
        >>> b["a"]
        Traceback (most recent call last):
            ...
        NameError: name 'a' is not defined at unknown position
        >>> b.setdefault("a",10)
        10
        >>> b["a"]
        10
        >>> b["a"]=11
        >>> b.setdefault("a",10)
        11
        """
        try:
            return self.__getitem__(name)
        except NameError as _:
            self.__setitem__(name,val)
        return val

    def set_substfile(self, filename, tp):
        """set substitution filename.

        Used for the "$subst", "$template" and "$pattern" commands.
        """
        if not isinstance(filename, str):
            raise EP.ParseException( \
                    self.posmsg("filename must be a string at",
                                pos=tp.start()))
        if filename == self.template:
            return
        self.template_parselist_cache= None
        self.template_path= None
        self.template= filename

    def substfile_parselist(self, include_paths):
        """return the parselist for a given file.

        This method manages a cache of that assigns filenames to parselists.
        This can speed things up if a file is included several times.
        """
        if self.template_parselist_cache is not None:
            return (self.template_path,self.template_parselist_cache)
        if self.template is None:
            raise ValueError(self.posmsg("substitition file name missing at"))
        self.template_path= find_file(self.template, include_paths)
        if self.template_path is None:
            raise ValueError(self.posmsg("file \"%s\" not found in" % \
                               self.template))
        else:
            self.template_parselist_cache= parseFile(self.template_path, False)
        return (self.template_path,self.template_parselist_cache)

    def export_symbols(self, lst):
        """appends items to the export_symbols list.

        This list is used by the pop() method in order to copy values from the
        current global variable dictionary to the global variable dictionary of
        the previous block.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> b.export_symbols(["a","b"])
        >>> b.exported_syms
        ['a', 'b']
        >>> b.export_symbols(["d","e"])
        >>> b.exported_syms
        ['a', 'b', 'd', 'e']
        """
        self.exported_syms.extend(lst)

    def extend(self, lst):
        """adds items to the list of expander functions or variables.

        Here is an example:

        >>> a=1
        >>> b=2
        >>> def t(x):
        ...   return x+1
        ...
        >>> block= Block(parse_list=[],external_definitions=globals())
        >>> block.extend(["a","b","t"])
        >>> _pr_set(block.direct_vars)
        set(['a', 'b'])
        >>> _pr_set(block.direct_funcs)
        set(['t'])
        """
        for elm in lst:
            obj= self.globals_[elm]
            if inspect.isbuiltin(obj):
                self.direct_funcs.add(elm)
                continue
            if inspect.isfunction(obj):
                self.direct_funcs.add(elm)
                continue
            # assume elm to be a variable:
            self.direct_vars.add(elm)

    def set_indent(self, value):
        """sets the indent."""
        self.indent= value

    def get_indent(self):
        """gets the indent."""
        return self.indent

    def format_text(self, text, indent_start):
        """currently does indent one or more lines."""
        if text=="":
            return text
        if self.indent<=0:
            return text
        ind= " "*self.indent
        cut= False
        if text.endswith(os.linesep):
            cut= True
        text= text.replace(os.linesep, os.linesep+ind)
        if cut:
            text= text[0:-self.indent]
        if not indent_start:
            return text
        else:
            return ind+text

    def add_macro(self, name, macro_block):
        """add a new macro block.
        """
        if not isinstance(macro_block, MacBlock):
            raise AssertionError("wrong type: %s has type %s" % \
                                 (name, type(macro_block)))
        self.macros[name]= macro_block

    def pop(self):
        """removes the current block and returns the previous one.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> b["a"]=1
        >>> b["b"]=2
        >>> b= Block(b,True)
        >>> b["a"]=10
        >>> b["b"]=20
        >>> b.export_symbols(["a"])
        >>> b= b.pop()
        >>> b["a"]
        10
        >>> b["b"]
        2
        """
        if self.previous is None:
            raise AssertionError(self.posmsg("block underflow (assertion) at"))
        if self.new_scope:
            old= self.previous.globals_
            for elm in self.exported_syms:
                old[elm]= self.globals_[elm]
        # set the lst_pos in the parent Block:
        if not self.new_parse_list:
            self.previous.lst_pos= self.lst_pos
        return self.previous

    def __str__(self):
        """returns a string representation of the block.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> print(b)
        Block{
            has parent          : False
            filename            = None
            template            = None
            template_path       = None
            has template_parselist_cache: False
            exported_syms       = []
            direct_vars         = set([])
            direct_funcs        = set([])
            macros              : []
            new_scope           = False
            skip                = False
            indent              = 0
            new_parse_list      = True
            lst_pos             = -1
            start_pos           = 0
        }
        >>> b= Block(b,True)
        >>> print(b)
        Block{
            has parent          : True
            filename            = None
            template            = None
            template_path       = None
            has template_parselist_cache: False
            exported_syms       = []
            direct_vars         = set([])
            direct_funcs        = set([])
            macros              : []
            new_scope           = True
            skip                = False
            indent              = 0
            new_parse_list      = False
            lst_pos             = -1
            start_pos           = 0
        }
        """
        lst=["%s{" % "Block"]
        lst.extend(self._strlist())
        lst.append("}")
        return "\n".join(lst)

    def get_block_list(self):
        """returns all blocks of the list.

        The list is returned with the oldest block first.
        """
        lst=[]
        block= self
        while block is not None:
            lst.append(block)
            block= block.previous
        lst.reverse()
        return lst

    def str_block_list(self):
        """returns a string representation of all blocks in the list.

        The list is returned with the oldest block first.
        """
        return [str(elm) for elm in self.get_block_list()]

    def print_block_list(self):
        """print all blocks in the list.

        The list is returned with the oldest block first.

        Here is an example:

        >>> b= Block(parse_list=[])
        >>> print(b)
        Block{
            has parent          : False
            filename            = None
            template            = None
            template_path       = None
            has template_parselist_cache: False
            exported_syms       = []
            direct_vars         = set([])
            direct_funcs        = set([])
            macros              : []
            new_scope           = False
            skip                = False
            indent              = 0
            new_parse_list      = True
            lst_pos             = -1
            start_pos           = 0
        }
        >>> b= Block(b,True)
        >>> print(b)
        Block{
            has parent          : True
            filename            = None
            template            = None
            template_path       = None
            has template_parselist_cache: False
            exported_syms       = []
            direct_vars         = set([])
            direct_funcs        = set([])
            macros              : []
            new_scope           = True
            skip                = False
            indent              = 0
            new_parse_list      = False
            lst_pos             = -1
            start_pos           = 0
        }
        >>> b.print_block_list()
        Block{
            has parent          : False
            filename            = None
            template            = None
            template_path       = None
            has template_parselist_cache: False
            exported_syms       = []
            direct_vars         = set([])
            direct_funcs        = set([])
            macros              : []
            new_scope           = False
            skip                = False
            indent              = 0
            new_parse_list      = True
            lst_pos             = -1
            start_pos           = 0
        }
        Block{
            has parent          : True
            filename            = None
            template            = None
            template_path       = None
            has template_parselist_cache: False
            exported_syms       = []
            direct_vars         = set([])
            direct_funcs        = set([])
            macros              : []
            new_scope           = True
            skip                = False
            indent              = 0
            new_parse_list      = False
            lst_pos             = -1
            start_pos           = 0
        }
        """
        print("\n".join(self.str_block_list()))

class IncludeBlock(Block):
    """implements a $include(filename) block.

    This block is simply a variable scope, so it is derived from Block where
    the constructor is called with new_scope=True.
    """
    def __init__(self,
                 previous= None,
                 new_scope= False,
                 filename= None,
                 include_paths=None):
        path= find_file(filename, include_paths)
        if path is None:
            parse_list= []
        else:
            parse_list= parseFile(path, False)
        Block.__init__(self, previous, new_scope,
                       path,
                       parse_list)
        if path is None:
            raise ValueError(self.posmsg("file \"%s\" not found in" % filename))
    def __str__(self):
        lst=["%s{" % "IncludeBlock"]
        lst.extend(self._strlist())
        lst.append("}")
        return "\n".join(lst)

class SubstBlock(Block):
    """implements a $subst(parameters) block.

    This block is simply a variable scope, so it is derived from Block where
    the constructor is called with new_scope=True.
    """
    def __init__(self,
                 previous= None,
                 filename= None,
                 include_paths=None,
                 external_definitions=None):
        (path,parse_list)= previous.substfile_parselist(include_paths)
        Block.__init__(self, previous, True, # always new scope
                       path,
                       parse_list,
                       external_definitions)
        if path is None:
            raise ValueError(self.posmsg("file \"%s\" not found in" % filename))
    def __str__(self):
        lst=["%s{" % "SubstBlock"]
        lst.extend(self._strlist())
        lst.append("}")
        return "\n".join(lst)

class PatternBlock(Block):
    """implements a $pattern(parameters) block.

    This block is simply a variable scope, so it is derived from Block where
    the constructor is called with new_scope=True.
    """
    def _strlist(self):
        lst= Block._strlist(self)
        self._append(lst, "heading")
        self._append(lst, "lines")
        self._append(lst, "curr_line")
        return lst
    def __init__(self,
                 previous= None,
                 filename= None,
                 include_paths=None,
                 heading=None,
                 lines=None):
        # pylint: disable=too-many-arguments
        (path,parse_list)= previous.substfile_parselist(include_paths)
        Block.__init__(self, previous, False, # no new scope
                       path, parse_list)
        if path is None:
            raise ValueError(self.posmsg("file \"%s\" not found in" % filename))
        if heading is None:
            heading= []
        if lines is None:
            lines= []
        self.heading= heading
        self.lines= lines
        if len(self.lines)<1:
            raise ValueError(self.posmsg("no instantiation data"))
        self.curr_line= -1
        self.def_vars()
    def pop(self):
        if not self.def_vars():
            return Block.pop(self)
        # reset position in file
        self.lst_pos= -1
        return self
    def def_vars(self):
        """define all the given variables in the PatternBlock."""
        self.curr_line+=1
        if self.curr_line >= len(self.lines):
            return False
        line= self.lines[self.curr_line]
        for i in range(0,len(self.heading)):
            self[self.heading[i]]= line[i]
        return True
    def __str__(self):
        lst=["%s{" % "PatternBlock"]
        lst.extend(self._strlist())
        lst.append("}")
        return "\n".join(lst)

class BeginBlock(Block):
    """implements a $begin .. $end block.

    This block is simply a variable scope, so it is derived from Block where
    the constructor is called with new_scope=True.
    """
    def __init__(self,
                 previous= None,
                 filename= None):
        Block.__init__(self, previous, True)
        self.skip= previous.skip
    def __str__(self):
        lst=["%s{" % "BeginBlock"]
        lst.extend(self._strlist())
        lst.append("}")
        return "\n".join(lst)

class IfBlock(Block):
    """implements a $if .. $else .. $endif block.

    An $if block never has a variable scope, so the base Block object is
    called with new_scope=False.
    """
    def _strlist(self):
        lst= Block._strlist(self)
        self._append(lst, "prev_skip")
        self._append(lst, "in_else_part")
        self._append(lst, "found")
        return lst
    def __init__(self,
                 previous= None,
                 condition= True):
        """constructs the $if block.

        condition is the boolean value of the $if condition.
        """
        Block.__init__(self, previous, False)
        self.prev_skip= previous.skip
        self.in_else_part= False
        if condition:
            self.found= True
            self.skip= self.prev_skip
        else:
            self.found= False
            self.skip= True
    def enter_elif(self, condition):
        """enter the "elif" part in the if..endif block."""
        if self.found:
            self.skip= True
        else:
            if condition:
                self.found= True
                self.skip= self.prev_skip
            else:
                self.skip= True
    def enter_else(self):
        """this should be called when $else is encountered.
        """
        if self.in_else_part:
            raise EP.ParseException( \
                    self.posmsg("one \"else\" too many at"))
        self.in_else_part= True
        if not self.found:
            self.skip= self.prev_skip
        else:
            self.skip= True
    def __str__(self):
        lst=["%s{" % "IfBlock"]
        lst.extend(self._strlist())
        lst.append("}")
        return "\n".join(lst)

class ForBlock(Block):
    """implements a $for .. $endfor block.
    """
    def _strlist(self):
        lst= Block._strlist(self)
        self._append(lst, "value_list")
        self._append(lst, "index")
        self._append(lst, "var_expr")
        self._append(lst, "jump_lst_pos")
        self._append(lst, "jump parse elm",self.parse_list[self.jump_lst_pos])
        return lst
    def __init__(self,
                 previous= None,
                 new_scope= False,
                 value_list=None,
                 var_expr=""):
        """constructor of the block.

        var_expr -- the expression that contains the loop variable or the
                    tuple with the loop variables.
        """
        Block.__init__(self, previous, new_scope)
        if value_list is None:
            value_list= []
        self.value_list= value_list
        self.index=0 # current index within self.value_list
        self.var_expr= var_expr
        self.jump_lst_pos= self.lst_pos
        if len(value_list)<=0:
            self.skip= True
        else:
            self.skip= previous.skip
    def set_loop_var(self):
        """set the loop variable to a new value."""
        if not self.skip:
            self.__setitem__(self.var_expr, self.value_list[self.index])
    def next_loop(self):
        """performs next loop.

        returns:
          True when the loop is not yet finished.
        """
        if self.skip:
            return False
        self.index+=1
        do_loop= self.index< len(self.value_list)
        if do_loop:
            self.lst_pos= self.jump_lst_pos
        return do_loop
    def __str__(self):
        lst=["%s{" % "ForBlock"]
        lst.extend(self._strlist())
        lst.append("}")
        return "\n".join(lst)

class WhileBlock(Block):
    """implements a $while .. $endwhile block.
    """
    def _strlist(self):
        lst= Block._strlist(self)
        self._append(lst, "while_expr")
        self._append(lst, "jump parse elm",self.parse_list[self.jump_lst_pos])
        return lst
    def __init__(self,
                 previous= None,
                 new_scope= False,
                 while_expr=""):
        """constructor of the block.

        while_expr -- the expression that contains the loop variable or the
                      tuple with the loop variables.
        """
        Block.__init__(self, previous, new_scope)
        self.while_expr= while_expr
        self.jump_lst_pos= self.lst_pos
        if while_expr=="":
            self.skip= True
        elif not self.eval_(self.while_expr):
            self.skip= True
        else:
            self.skip= previous.skip
    def next_loop(self):
        """performs next loop.

        returns:
          True when the loop is not yet finished.
        """
        if self.skip:
            return False
        do_loop= self.eval_(self.while_expr)
        if do_loop:
            self.lst_pos= self.jump_lst_pos
        return do_loop
    def __str__(self):
        lst=["%s{" % "WhileBlock"]
        lst.extend(self._strlist())
        lst.append("}")
        return "\n".join(lst)

class MacBlock(Block):
    """implements a $macro...$endmacro block.
    """
    def _strlist(self):
        lst= Block._strlist(self)
        self._append(lst, "parameter_list")
        self._append(lst, "is_declaration")
        return lst
    def __init__(self,
                 previous= None,
                 declaration_block= None,
                 parameter_list= None):
        """constructor of the block.

        parameter_list: list of parameters of the macro.
        """
        if declaration_block is None:
            # THIS is a macro declaration block
            Block.__init__(self, previous, new_scope= True)
            if parameter_list is None:
                parameter_list= []
            self.parameter_list= parameter_list
            self.is_declaration= True
            self.skip= True
        else:
            # THIS is a macro instantiation block
            Block.__init__(self, previous, new_scope= True)
            self.parse_list= declaration_block.parse_list
            self.parameter_list= declaration_block.parameter_list
            self.lst_pos= declaration_block.start_pos
            self.is_declaration= False

    def pop(self):
        """override pop() from base class.
        """
        if self.is_declaration:
            return Block.pop(self)
        old_pos= self.previous.lst_pos
        prev= Block.pop(self)
        prev.lst_pos= old_pos
        return prev

    def __str__(self):
        lst=["%s{" % "MacBlock"]
        lst.extend(self._strlist())
        lst.append("}")
        return "\n".join(lst)

def __pyexpander_helper(*args, **kwargs):
    """a helper function needed at runtime.

    This evaluates named arguments.
    """
    if len(args)==1:
        fn= args[0]
    elif len(args)>1:
        raise ValueError("only one unnamed argument is allowed")
    else:
        fn= None
    return(fn, kwargs)

def __pyexpander_helper2(**kwargs):
    """a helper function needed at runtime.

    This evaluates named arguments.
    """
    return kwargs

def __pyexpander_helper3(*args, **kwargs):
    """a helper function needed at runtime.

    This evaluates unnamed and named arguments.
    """
    return (args,kwargs)

# contains an indent at the end of the string
rx_ENDING_INDENT = re.compile(r'\n[\t ]+$')
def processToList(parse_list, filename=None,
                  external_definitions=None,
                  allow_nobracket_vars= False,
                  auto_continuation= False,
                  auto_indent= False,
                  include_paths= None,
                  auto_indent_python=False):
    """Expand a parse list to a list of strings.

    args:
        parse_list                 : A parse list created by parseString().
        filename (str)             : The filename, if given, is included in
                                     possible error messages.
        external_definitions (dict): A dict with items to import to the
                                     globals() dictionary.
        allow_nobracket_vars (bool): If True, allow variables in the form $VAR
                                     instead of $(VAR).
        auto_continuation (bool)   : If True, remove newline at the end of
                                     lines with a command. This works like
                                     having an '\' at the end of each line with
                                     a command.
        auto_indent (bool)         : If True, indent the contents of macros to
                                     the same level as the macro invocation.
        include_paths (list)       : A list of paths that are searched for the
                                     $include command. 
        auto_indent_python (bool)   : If True, indents the contents of python
                                     output to the same level

    returns a tuple containing:
        - The expanded text as a list of strings
        - The internal globals() dictionary.
    """
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    # pylint: disable=too-many-arguments
    # accept None for include_paths too:
    if include_paths is None:
        include_paths= []
    # prepend the cwd to the list of search paths:
    include_paths.insert(0,"")
    # The initial block:
    my_external_definitions= { "__pyexpander_helper": \
                                 globals()["__pyexpander_helper"],
                               "__pyexpander_helper2": \
                                 globals()["__pyexpander_helper2"],
                               "__pyexpander_helper3": \
                                 globals()["__pyexpander_helper3"],
                             }
    if external_definitions is not None:
        my_external_definitions.update(external_definitions)
    # only needed for allow_nobracket_vars==True:
    # python keywords MUST NOT be checked. E.g. function __getitem__ would
    # raise an exception when called with "else" since eval("else") is a syntax
    # error.
    keyword_checks= PURE_CMD_KEYWORDS - PY_KEYWORDS
    block= Block(filename= filename, parse_list=parse_list,
                 external_definitions= my_external_definitions)
    result= ResultText()
    # needed for tp_last:
    tp= None
    # pylint: disable=too-many-nested-blocks
    while True:
        # print "-" * 40
        # block.print_block_list()
        if not block.parse_loop():
            # no more data in the current block:
            if isinstance(block, IncludeBlock):
                # if current block is an IncludeBlock, go back to previous
                # block:
                block= block.pop()
                continue
            elif isinstance(block, SubstBlock):
                # if current block is a SubstBlock, go back to previous block:
                block= block.pop()
                continue
            elif isinstance(block, PatternBlock):
                # if current block is a PatternBlock, go back to previous
                # block:
                block= block.pop()
                continue
            else:
                # end of data, leave the loop:
                break
        # get the current parse element (base class: ParsedItem)
        tp_last= tp
        tp= block.parse_elm()
        # print("POS %3d: " % block.lst_pos, "tp: ",str(tp))
        if isinstance(tp, EP.ParsedComment):
            # comments are ignored:
            continue
        if isinstance(tp, EP.ParsedLiteral):
            # literals are only taken if skip mode is off:
            if not block.skip:
                st_= tp.string()
                if auto_continuation and st_.startswith(os.linesep):
                    if isinstance(tp_last, EP.ParsedCommand) or \
                       isinstance(tp_last, EP.ParsedPureCommand):
                        st_= st_[LINESEP_LEN:]
                # if current column is 1, we must do an initial indent:
                st_= block.format_text(st_, result.column()==1)
                result.append(st_)
            continue
        if isinstance(tp, EP.ParsedVar):
            # if skip mode is off, insert the current value of the variable.
            # The current block can be used like a dict in order to get values
            # of variables:
            if not block.skip:
                # checks if there is an indent behind the variable as a literal
                if auto_indent_python and isinstance(tp_last, EP.ParsedLiteral):
                    search = re.search(rx_ENDING_INDENT, tp_last.string())
                    if search is not None:
                        # group() contains a newline at the beginning of the string
                        indent = search.group()
                        string = str(block[tp.string()])
                        result.append(string.replace("\n", indent))
                        continue
                result.append(str(block[tp.string()]))
            continue
        if isinstance(tp, EP.ParsedEval):
            # if skip mode is off, evaluate the eval expression,
            # convert it to a string and insert the result:
            if not block.skip:
                if auto_indent_python and isinstance(tp_last, EP.ParsedLiteral):
                    search = re.search(rx_ENDING_INDENT, tp_last.string())
                    if search is not None:
                        # group() contains a newline at the beginning of the string
                        indent = search.group()
                        string = block.str_eval(tp.string())
                        result.append(string.replace("\n", indent))
                        continue
                # wtf?
                text = block.str_eval(tp.string())
                indented_text = block.format_text(text, result.column()==1)
                # print(repr(text), "->", repr(indented_text))
                result.append(indented_text)

            continue
        if isinstance(tp, EP.ParsedCommand):
            # if ParsedItem is a ParsedCommand:
            if tp.ident=="py":
                # $py(...) :
                # execute the string given within the brackets:
                if not block.skip:
                    block.exec_(tp.args())
            elif tp.ident=="include" or tp.ident=="include_begin":
                if not block.skip:
                    # $include(...) or $include_begin(...) :
                    # evaluate the filename of the file to include:
                    filename= block.str_eval(tp.args())
                    if not isinstance(filename, str):
                        raise EP.ParseException( \
                                block.posmsg("filename must be a string at", \
                                              pos=tp.start()))
                    # pylint: disable= redefined-variable-type
                    # create an instance of an IncludeBlock:
                    block= IncludeBlock(previous= block,
                                        new_scope= (tp.ident=="include_begin"),
                                        filename= filename,
                                        include_paths= include_paths)
            elif tp.ident=="template":
                if not block.skip:
                    # $template(...) :
                    # evaluate the filename of substfile:
                    filename= block.str_eval(tp.args())
                    # remember this filename in the current block object:
                    block.set_substfile(filename, tp)
            elif tp.ident=="subst":
                if not block.skip:
                    # $subst(...) :
                    # evaluate all the named arguments:
                    args= block.eval_("__pyexpander_helper(%s)" % tp.args())
                    # first argument may be the name of the substfile,
                    # this is undocumented and usually not used:
                    fn= args[0]
                    if fn is not None:
                        block.set_substfile(fn, tp)
                    # create an instance of a SubstBlock:
                    block= SubstBlock(previous= block,
                                      filename= block.template,
                                      include_paths= include_paths,
                                      external_definitions= args[1])
            elif tp.ident=="pattern":
                if not block.skip:
                    # $pattern(...) :
                    # create a tuple of all arguments:
                    args= block.eval_("(%s)" % tp.args())
                    firstindex= 0
                    # first argument may be the name of the substfile,
                    # this is undocumented and usually not used:
                    if isinstance(args[0], str):
                        block.set_substfile(fn, tp)
                        firstindex= 1
                    block= PatternBlock(previous= block,
                                        filename= block.template,
                                        include_paths= include_paths,
                                        heading= args[firstindex],
                                        lines= args[firstindex+1:])
            elif tp.ident=="default":
                if not block.skip:
                    # $default(...) :
                    # evaluate all the named arguments:
                    args= block.eval_("__pyexpander_helper2(%s)" % tp.args())
                    # set these defaults in the current block:
                    for (k,v) in list(args.items()):
                        block.setdefault(k, v)
            elif tp.ident=="if":
                # $if(...) :
                # evaluate the condition:
                if not block.skip:
                    condition= block.eval_(tp.args())
                else:
                    # fake a condition:
                    condition= True
                # create an instance of an IfBlock:
                block= IfBlock(previous= block,
                               condition= condition)
            elif tp.ident=="elif":
                # elif(...) :
                # current block must be an IfBlock:
                if not isinstance(block,IfBlock):
                    raise EP.ParseException( \
                            block.posmsg("unmatched elif at", \
                                          pos=tp.start()))
                # pylint: disable=no-member
                # evaluate the condition:
                if not block.prev_skip:
                    condition= block.eval_(tp.args())
                else:
                    # fake a condition:
                    condition= False
                # enter the "elif" part of the if block by
                # calling enter_elif:
                block.enter_elif(condition)
            elif tp.ident=="for" or tp.ident=="for_begin":
                # $for(...) or $for_begin(...) :
                # assume the the parameters form a valid list comprehension
                e= None
                try:
                    # try to parse the arguments of $for():
                    for_parts= EP.scanPyIn(tp.args())
                except EP.ParseException as _e:
                    e= _e
                if e is not None:
                    # pylint: disable=raising-non-exception
                    raise e.__class__( \
                            block.posmsg("error in %s command at" % tp.ident, \
                                         pos= tp.start()))
                # create a list of loop items by using pythons
                # list comprehension mechanism:
                if not block.skip:
                    for_list= block.eval_("list(%s)" % for_parts[2])
                else:
                    for_list= []
                # create an instance of a ForBlock:
                block= ForBlock(previous= block,
                                new_scope= (tp.ident=="for_begin"),
                                value_list= for_list,
                                var_expr= for_parts[0])
                block.set_loop_var()
            elif tp.ident=="while" or tp.ident=="while_begin":
                # $while(...) or $while_begin(...) :
                # create an instance of a WhileBlock:
                if not block.skip:
                    expr= tp.args()
                else:
                    expr= ""
                block= WhileBlock(previous= block,
                                  new_scope= (tp.ident=="while_begin"),
                                  while_expr= expr)
            elif tp.ident=="macro":
                # $macro(...)
                try:
                    # try to parse the arguments of $macro():
                    identifiers= EP.scanPyIdentList(tp.args())
                except EP.ParseException as _:
                    raise EP.ParseException(\
                          block.posmsg("error in \"macro\" command at", \
                                       pos= tp.start()))
                if len(identifiers)<=0:
                    raise EP.ParseException( \
                          block.posmsg("error in \"macro\" command at", \
                                       pos= tp.start()))
                e= None
                try:
                    keyword_check(identifiers[0:1])
                except ValueError as _e:
                    # pylint: disable=redefined-variable-type
                    e= _e
                if e is not None:
                    raise EP.ParseException( \
                          block.posmsg(("error in \"macro\" command, "
                                        "%s at") % e, \
                                       pos= tp.start()))
                block= MacBlock(previous= block,
                                parameter_list= identifiers[1:])
                block.previous.add_macro(identifiers[0], block)
            elif tp.ident=="nonlocal":
                if not block.skip:
                    # $nonlocal(...) :
                    e= None
                    try:
                        # try to parse the arguments of $nonlocal():
                        identifiers= EP.scanPyIdentList(tp.args())
                    except EP.ParseException as _e:
                        e= _e
                    if e is not None:
                        # pylint: disable=raising-non-exception
                        raise e.__class__( \
                              block.posmsg("error in \"nonlocal\" command at", \
                                           pos= tp.start()))
                    # mark them in the current block as exported symbols:
                    block.export_symbols(identifiers)
            elif tp.ident=="extend" or tp.ident=="extend_expr":
                if not block.skip:
                    if tp.ident=="extend_expr":
                        # $extend_expr(...) :
                        expr_= block.eval_(tp.args())
                        e= None
                        try:
                            identifiers= [str(elm) for elm in expr_]
                        except TypeError as _e:
                            e= _e
                        if e is not None:
                            raise e.__class__("%s at %s" % \
                                              (str(e), block.posmsg()))
                    else:
                        # $extend(...) :
                        e= None
                        try:
                            # try to parse the arguments of $extend():
                            identifiers= EP.scanPyIdentList(tp.args())
                        except EP.ParseException as _e:
                            e= _e
                        if e is not None:
                            # pylint: disable=raising-non-exception
                            raise e.__class__( \
                                  block.posmsg("error in \"extend\" command at", \
                                               pos= tp.start()))
                    e= None
                    try:
                        keyword_check(identifiers)
                    except ValueError as _e:
                        e= _e
                    if e is not None:
                        raise EP.ParseException( \
                              block.posmsg(("error in \"extend\" command, "
                                            "%s at") % e, \
                                           pos= tp.start()))
                    # mark them in the current block as "extended" identifiers:
                    block.extend(identifiers)
            elif tp.ident in block.direct_funcs:
                if not block.skip:
                    # $user-function-extended(...)
                    # apply the function directly:
                    result.append(block.str_eval("%s(%s)" % \
                                  (tp.ident,tp.args())))

            elif tp.ident in block.macros:
                if not block.skip:
                    # $macro(args)
                    block= MacBlock(previous= block,
                                    declaration_block= block.macros[tp.ident],
                                    parameter_list= identifiers[1:])
                    if auto_indent:
                        # column of the '$' sign:
                        col_= tp.rowcol()[1]-len(tp.ident)-2
                        block.set_indent(block.get_indent()+(col_-1))
                        # print(block)
                        # print(tp)
                        # print(repr(tp.args()))
                    (args,kwargs)= block.eval_("__pyexpander_helper3(%s)" % \
                                               tp.args())

                    if len(args)>len(block.parameter_list):
                        raise EP.ParseException( \
                                block.posmsg(("too many parameters at "
                                              "instantiation of macro "
                                              "\"%s\" at") % tp.ident,
                                             pos= tp.start()))

                    for (i,name) in enumerate(block.parameter_list):
                        if i<len(args):
                            # unnamed parameter
                            if name in kwargs:
                                raise EP.ParseException( \
                                        block.posmsg(("multiple values for "
                                                      "keyword argument "
                                                      "\"%s\" in macro "
                                                      "\"%s\" at") % \
                                                     (name, tp.ident),
                                                     pos= tp.start()))
                            block.__setitem__(name, args[i])
                            continue
                        if not name in kwargs:
                            raise EP.ParseException( \
                                    block.posmsg(("no value for argument "
                                                  "\"%s\" in macro \"%s\" "
                                                  "at") % \
                                                 (name, tp.ident),
                                                 pos= tp.start()))
                        block.__setitem__(name, kwargs[name])
            else:
                # everything else is a ParseException, this shouldn't happen
                # since all ParsedItem objects that the expanderparser can
                # create should be handled here.
                raise EP.ParseException( \
                        block.posmsg("unknown command \"%s\" at" % tp.ident, \
                                      pos= tp.start()))
            continue
        if isinstance(tp, EP.ParsedPureCommand):
            # a "pure" command, a command without arguments:
            ident= tp.string()
            if allow_nobracket_vars and (ident in keyword_checks):
                # an extra check for keyword conflicts
                do_warn= False
                try:
                    block.__getitem__(ident)
                    do_warn= True
                except NameError as _:
                    pass
                if do_warn:
                    sys.stderr.write(block.posmsg(("warning, variable '%s' "
                                                   "shadowed by pyexpander "
                                                   "keyword (This warning "
                                                   "only printed once per "
                                                   "keyword) at" % ident),
                                                  pos= tp.start())+"\n")
                    keyword_checks.remove(ident)
            if ident=="else":
                # $else :
                # current block must be an IfBlock:
                if not isinstance(block,IfBlock):
                    raise EP.ParseException( \
                            block.posmsg("unmatched else at", \
                                          pos= tp.start()))
                # enter the "else" part of the if block by
                # calling enter_else:
                block.enter_else()
            elif ident=="endif":
                # $endif
                # current block must be an IfBlock:
                if not isinstance(block,IfBlock):
                    raise EP.ParseException( \
                            block.posmsg("unmatched endif at", \
                                          pos= tp.start()))
                # go back to previous block:
                block= block.pop()
            elif ident=="endfor":
                # $endfor
                # current block must be a ForBlock:
                if not isinstance(block,ForBlock):
                    raise EP.ParseException( \
                            block.posmsg("unmatched endfor at", \
                                          pos= tp.start()))
                # test if we have to perform the loop block again:
                if not block.next_loop():
                    # no further loops, go back to the previous block:
                    block= block.pop()
                else:
                    # further loops, give the loop variable a new value:
                    block.set_loop_var()
            elif ident=="endwhile":
                # $endwhile
                # current block must be a WhileBlock:
                if not isinstance(block,WhileBlock):
                    raise EP.ParseException( \
                            block.posmsg("unmatched endwhile at", \
                                          pos= tp.start()))
                # test if we have to loop again. next_loop also resets
                # the position, if we have to loop again:
                if not block.next_loop():
                    # if loop condition is False, go back to previous block,
                    block= block.pop()
            elif ident=="endmacro":
                # $endmacro
                # current block must be a MacBlock:
                if not isinstance(block,MacBlock):
                    raise EP.ParseException( \
                            block.posmsg("unmatched endmacro at", \
                                          pos= tp.start()))
                block= block.pop()
            elif ident=="begin":
                # $begin
                # create an instance of a BeginBlock:
                block= BeginBlock(previous= block)
            elif ident=="end":
                # $end
                # current block must be a BeginBlock:
                if not isinstance(block,BeginBlock):
                    raise EP.ParseException( \
                            block.posmsg("unmatched end at", \
                                          pos= tp.start()))
                # go back to previous block:
                block= block.pop()
            elif ident in block.direct_vars:
                # $user-variable-extended
                # if skip mode is off, insert the current value of the
                # variable. The current block can be used like a dict in order
                # to get values of variables:
                if not block.skip:
                    result.append(str(block.__getitem__(ident)))
            else:
                # if we are not in nobracket_vars mode, we have an
                # unknown command without arguments here:
                if not allow_nobracket_vars:
                    raise EP.ParseException( \
                            block.posmsg("unknown command \"%s\" at" % ident, \
                                          pos= tp.start()))
                else:
                    # if skip mode is off, insert the current value of the
                    # variable. The current block can be used like a dict in
                    # order to get values of variables:
                    if not block.skip:
                        result.append(str(block.__getitem__(ident)))
            continue
    if block.previous is not None:
        raise EP.ParseException(block.posmsg("unclosed block at"))
    return (result.list_(),block.globals_)

def processToPrint(parse_list, **kwargs):
    """Gets a parse list, expand the text in it and print it.

    args: see `processToList`

    returns:
        The internal globals() dictionary.
    """
    # pylint: disable=too-many-arguments
    # debug:
    # for elm in parse_list:
    #     print elm
    (result,exp_globals)= processToList(parse_list, **kwargs)
    print("".join(result), end='')
    return exp_globals

def expandToStr(st, **kwargs):
    """Get a string, expand the text in it and return it as a string.

    args: see `processToList`

    returns a tuple containing:
        - The expanded text as a single string.
        - The internal globals() dictionary.
    """
    # pylint: disable=too-many-arguments
    (result,exp_globals)= processToList(parseString(st), **kwargs)
    return ("".join(result), exp_globals)

def expand(st, **kwargs):
    """Get a string, expand the text in it and print it.

    args: see `processToList`

    returns:
        The internal globals() dictionary.
    """
    # pylint: disable=too-many-arguments
    return processToPrint(parseString(st), **kwargs)

def expandFile(filename, no_stdin_warning=False, **kwargs):
    """Get a filename, expand the text in it and print it.

    args: (see `processToList`)
        no_stdin_warning (bool)    : If True, print short message on stderr
                                     when the program is waiting on input from
                                     stdin.

    returns:
        The internal globals() dictionary.
    """
    # pylint: disable=too-many-arguments
    return processToPrint(parseFile(filename, no_stdin_warning), **kwargs)

def _test():
    """perform the doctest tests."""
    import doctest
    print("testing...")
    doctest.testmod()
    print("done")

if __name__ == "__main__":
    _test()



if __name__ == "__main__":
    import sys
    sys.path.append("..")

import fena_pyexpander.lib as pyexpander

with open("test.txt") as file:
    text = file.read()

output = pyexpander.expandToStr(text, filename="test.txt", auto_indent=True, auto_indent_python=True)[0]
print(output)


